"""
A client interface for Sentinel Hub Catalog API
"""
from .config import SHConfig
from .data_collections import DataCollection, OrbitDirection
from .geometry import Geometry, CRS
from .download.sentinelhub_client import SentinelHubDownloadClient
from .sh_utils import FeatureIterator, remove_undefined
from .time_utils import parse_time_interval, serialize_time, parse_time


class SentinelHubCatalog:
    """ The main class for interacting with Sentinel Hub Catalog API

    For more details about certain endpoints and parameters check
    `Catalog API documentation <https://docs.sentinel-hub.com/api/latest/api/catalog>`_.
    """

    def __init__(self, base_url=None, config=None):
        """
        :param base_url: A base URL of Sentinel Hub service specifying which service deployment to use. By default the
            one in config object will be used.
        :type base_url: str or None
        :param config: A configuration object with required parameters `sh_client_id`, `sh_client_secret`, and
            `sh_base_url` which will be used of authentication.
        :type config: SHConfig or None
        """
        self.config = config or SHConfig()

        base_url = base_url or self.config.sh_base_url
        base_url = base_url.rstrip('/')
        self.catalog_url = f'{base_url}/api/v1/catalog'

        self.client = SentinelHubDownloadClient(config=self.config)

    def get_info(self):
        """ Provides the main information that define Sentinel Hub Catalog API

        `Catalog API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/getLandingPage>`__

        :return: A service payload with information
        :rtype: dict
        """
        return self.client.get_json(self.catalog_url)

    def get_conformance(self):
        """ Get information about specifications that this API conforms to

        `Catalog API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#operation/getConformanceDeclaration>`__

        :return: A service payload with information
        :rtype: dict
        """
        url = f'{self.catalog_url}/conformance'
        return self.client.get_json(url)

    def get_collections(self):
        """ Provides a list of collections that are available to a user

        `Catalog API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/getCollections>`__

        :return: A list of collections with information
        :rtype: list(dict)
        """
        url = f'{self.catalog_url}/collections'
        return self.client.get_json(url, use_session=True)['collections']

    def get_collection(self, collection):
        """ Provides information about given collection

        `Catalog API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/describeCollection>`__

        :param collection: A data collection object or a collection ID
        :type collection: DataCollection or str
        :return: Information about a collection
        :rtype: dict
        """
        collection_id = self._parse_collection_id(collection)
        url = f'{self.catalog_url}/collections/{collection_id}'
        return self.client.get_json(url, use_session=True)

    def get_feature(self, collection, feature_id):
        """ Provides information about a single feature in a collection

        `Catalog API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/getFeature>`__

        :param collection: A data collection object or a collection ID
        :type collection: DataCollection or str
        :param feature_id: A feature ID
        :type feature_id: str
        :return: Information about a feature in a collection
        :rtype: dict
        """
        collection_id = self._parse_collection_id(collection)
        url = f'{self.catalog_url}/collections/{collection_id}/items/{feature_id}'
        return self.client.get_json(url, use_session=True)

    def search(self, collection, *, time, bbox=None, geometry=None, ids=None, query=None, fields=None, distinct=None,
               limit=100, **kwargs):
        """ Catalog STAC search

        `Catalog API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/postSearchSTAC>`__

        :param collection: A data collection object or a collection ID
        :type collection: DataCollection or str
        :param time: A time interval or a single time. It can either be a string in form  YYYY-MM-DDThh:mm:ss or
            YYYY-MM-DD or a datetime object
        :type time: (str, str) or (datetime, datetime) or str or datetime
        :param bbox: A search bounding box, it will always be reprojected to WGS 84 before being sent to the service.
            Re-projection will be done with BBox.transform_bounds method which can produce a slightly larger bounding
            box. If that is a problem then transform a BBox object into a Geometry object and search with geometry
            parameter instead.
        :type bbox: BBox
        :param geometry: A search geometry, it will always reprojected to WGS 84 before being sent to the service.
            This parameter is defined with parameter `intersects` at the service.
        :type geometry: Geometry
        :param ids: A list of feature ids as defined in service documentation
        :type ids: list(str)
        :param query: A STAC query described in Catalog API documentation
        :type query: dict
        :param fields: A dictionary of fields to include or exclude described in Catalog API documentation
        :type fields: dict
        :param distinct: A special query attribute described in Catalog API documentation
        :type distinct: str
        :param limit: A number of results to return per each request. At the end iterator will always provide all
            results the difference is only in how many requests it will have to make in the background.
        :type limit: int
        :param kwargs: Any other parameters that will be passed directly to the service
        """
        url = f'{self.catalog_url}/search'

        collection_id = self._parse_collection_id(collection)
        start_time, end_time = serialize_time(parse_time_interval(time, allow_undefined=True), use_tz=True)

        if bbox and bbox.crs is not CRS.WGS84:
            bbox = bbox.transform_bounds(CRS.WGS84)
        if geometry and geometry.crs is not CRS.WGS84:
            geometry = geometry.transform(CRS.WGS84)

        _query = self._get_data_collection_filters(collection)
        if query:
            _query.update(query)

        payload = remove_undefined({
            'collections': [collection_id],
            'datetime': f'{start_time}/{end_time}' if time else None,
            'bbox': list(bbox) if bbox else None,
            'intersects': geometry.geojson if geometry else None,
            'ids': ids,
            'query': _query,
            'fields': fields,
            'distinct': distinct,
            'limit': limit,
            **kwargs
        })

        return CatalogSearchIterator(self.client, url, payload)

    @staticmethod
    def _parse_collection_id(collection):
        """ Extracts catalog collection id from an object defining a collection
        """
        if isinstance(collection, DataCollection):
            return collection.catalog_id
        if isinstance(collection, str):
            return collection
        raise ValueError(f'Expected either a DataCollection object or a collection id string, got {collection}')

    @staticmethod
    def _get_data_collection_filters(data_collection):
        """ Builds a dictionary of query filters for catalog API from a data collection definition
        """
        filters = {}

        if data_collection.swath_mode:
            filters['sar:instrument_mode'] = {'eq': data_collection.swath_mode.upper()}

        if data_collection.polarization:
            filters['polarization'] = {'eq': data_collection.polarization.upper()}

        if data_collection.resolution:
            filters['resolution'] = {'eq': data_collection.resolution.upper()}

        if data_collection.orbit_direction and data_collection.orbit_direction.upper() != OrbitDirection.BOTH:
            filters['sat:orbit_state'] = {'eq': data_collection.orbit_direction.lower()}

        if data_collection.timeliness:
            filters['timeliness'] = {'eq': data_collection.timeliness}

        return filters


class CatalogSearchIterator(FeatureIterator):
    """ Searches a catalog with a given query and provides results
    """

    def _fetch_features(self):
        """ Collects more results from the service
        """
        payload = remove_undefined({
            **self.params,
            'next': self.next
        })

        results = self.client.get_json(self.url, post_values=payload, use_session=True)

        self.next = results['context'].get('next')
        new_features = results['features']
        self.finished = self.next is None or not new_features

        return new_features

    def get_timestamps(self):
        """ Provides features timestamps

        :return: A list of sensing times
        :rtype: list(datetime.datetime)
        """
        return [parse_time(feature['properties']['datetime']) for feature in self]

    def get_geometries(self):
        """ Provides features geometries

        :return: A list of geometry objects with CRS
        :rtype: list(Geometry)
        """
        return [Geometry.from_geojson(feature['geometry']) for feature in self]

    def get_ids(self):
        """ Provides features IDs

        :return: A list of IDs
        :rtype: list(str)
        """
        return [feature['id'] for feature in self]



"""
Module with useful time/date functions
"""
import datetime as dt
import warnings

import dateutil.parser
import dateutil.tz

from .exceptions import SHDeprecationWarning


def get_dates_in_range(start_date, end_date):
    """ Get all dates within input start and end date in ISO 8601 format

    :param start_date: start date in ISO 8601 format
    :type start_date: str
    :param end_date: end date in ISO 8601 format
    :type end_date: str
    :return: list of dates between start_date and end_date in ISO 8601 format
    :rtype: list of str
    """
    warnings.warn('This function will be removed in the next version', category=SHDeprecationWarning)
    start_dt = iso_to_datetime(start_date)
    end_dt = iso_to_datetime(end_date)
    num_days = int((end_dt - start_dt).days)
    return [datetime_to_iso(start_dt + dt.timedelta(i)) for i in range(num_days + 1)]


def next_date(date):
    """ Get date of day after input date in ISO 8601 format

    For instance, if input date is ``'2017-03-12'``, the function returns ``'2017-03-13'``

    :param date: input date in ISO 8601 format
    :type date: str
    :return: date after input date in ISO 8601 format
    :rtype: str
    """
    warnings.warn('This function will be removed in the next version', category=SHDeprecationWarning)
    dtm = iso_to_datetime(date)
    return datetime_to_iso(dtm + dt.timedelta(1))


def prev_date(date):
    """ Get date of day previous to input date in ISO 8601 format

    For instance, if input date is ``'2017-03-12'``, the function returns ``'2017-03-11'``

    :param date: input date in ISO 8601 format
    :type date: str
    :return: date previous to input date in ISO 8601 format
    :rtype: str
    """
    warnings.warn('This function will be removed in the next version', category=SHDeprecationWarning)
    dtm = iso_to_datetime(date)
    return datetime_to_iso(dtm - dt.timedelta(1))


def iso_to_datetime(date):
    """ Convert ISO 8601 time format to datetime format

    This function converts a date in ISO format, e.g. ``2017-09-14`` to a `datetime` instance, e.g.
    ``datetime.datetime(2017,9,14,0,0)``

    :param date: date in ISO 8601 format
    :type date: str
    :return: datetime instance
    :rtype: datetime
    """
    warnings.warn('This function will be removed in the next version', category=SHDeprecationWarning)
    chunks = list(map(int, date.split('T')[0].split('-')))
    return dt.datetime(chunks[0], chunks[1], chunks[2])


def datetime_to_iso(date, only_date=True):
    """ Convert datetime format to ISO 8601 time format

    This function converts a date in datetime instance, e.g. ``datetime.datetime(2017,9,14,0,0)`` to ISO format,
    e.g. ``2017-09-14``

    :param date: datetime instance to convert
    :type date: datetime
    :param only_date: whether to return date only or also time information. Default is `True`
    :type only_date: bool
    :return: date in ISO 8601 format
    :rtype: str
    """
    warnings.warn('This function will be removed in the next version', category=SHDeprecationWarning)
    if only_date:
        return date.isoformat().split('T')[0]
    return date.isoformat()


def get_current_date():
    """ Get current date in ISO 8601 format

    :return: current date in ISO 8601 format
    :rtype: str
    """
    warnings.warn('This function will be removed in the next version', category=SHDeprecationWarning)
    return datetime_to_iso(dt.datetime.now())


def is_valid_time(time):
    """ Check if input string represents a valid time/date stamp

    :param time: a string containing a time/date stamp
    :type time: str
    :return: `True` is string is a valid time/date stamp, `False` otherwise
    :rtype: bool
    """
    try:
        dateutil.parser.parse(time)
        return True
    except dateutil.parser.ParserError:
        return False


def parse_time(time_input, *, force_datetime=False, allow_undefined=False, **kwargs):
    """ Parse input time/date string

    :param time_input: time/date to parse
    :type time_input: str or datetime.date or datetime.datetime
    :param force_datetime: If True it will always return datetime.datetime object, if False it can also return only
        datetime.date object if only date is provided as input.
    :type force_datetime: bool
    :param allow_undefined: Flag to allow parsing None or '..' into None
    :param allow_undefined: bool (default is False)
    :param kwargs: Any keyword arguments to be passed to `dateutil.parser.parse`. Example: `ignoretz=True`
    :return: A datetime object
    :rtype: datetime.datetime or datetime.date
    """

    if allow_undefined and time_input in [None, '..']:
        return None

    if isinstance(time_input, dt.date):
        if force_datetime and not isinstance(time_input, dt.datetime):
            return date_to_datetime(time_input)

        if kwargs.get('ignoretz') and isinstance(time_input, dt.datetime):
            return time_input.replace(tzinfo=None)

        return time_input

    time = dateutil.parser.parse(time_input, **kwargs)
    if force_datetime or len(time_input) > 10:  # This check is not very accurate but it works for iso format
        return time
    return time.date()


def parse_time_interval(time, allow_undefined=False, **kwargs):
    """ Parse input into an interval of two times, specifying start and end time, into datetime objects.

    The input time can have the following formats, which will be parsed as:

    * `YYYY-MM-DD` -> `[YYYY-MM-DD:T00:00:00, YYYY-MM-DD:T23:59:59]`
    * `YYYY-MM-DDThh:mm:ss` -> `[YYYY-MM-DDThh:mm:ss, YYYY-MM-DDThh:mm:ss]`
    * list or tuple of two dates in form `YYYY-MM-DD` -> `[YYYY-MM-DDT00:00:00, YYYY-MM-DDT23:59:59]`
    * list or tuple of two dates in form `YYYY-MM-DDThh:mm:ss` -> `[YYYY-MM-DDThh:mm:ss, YYYY-MM-DDThh:mm:ss]`

    All input times can also be specified as `datetime` objects. Instances of `datetime.date` will be treated as
    `YYYY-MM-DD` and instance of `datetime.datetime` will be treated as `YYYY-MM-DDThh:mm:ss`.

    :param allow_undefined: Boolean flag controls if None or '..' are allowed
    :param allow_undefined: bool
    :param time: An input time
    :type time: str or datetime.datetime or (str, str) or (datetime.datetime, datetime.datetime)
    :return: interval of start and end date of the form `YYYY-MM-DDThh:mm:ss`
    :rtype: (datetime.datetime, datetime.datetime)
    :raises: ValueError
    """
    if allow_undefined and time in [None, '..']:
        date_interval = None, None
    elif isinstance(time, (str, dt.date)):
        parsed_time = parse_time(time, **kwargs)
        date_interval = parsed_time, parsed_time
    elif isinstance(time, (tuple, list)) and len(time) == 2:
        date_interval = parse_time(time[0], allow_undefined=allow_undefined, **kwargs), \
                        parse_time(time[1], allow_undefined=allow_undefined, **kwargs)
    else:
        raise ValueError('Time must be a string/datetime object or tuple/list of 2 strings/datetime objects')

    start_time, end_time = date_interval
    if not isinstance(start_time, dt.datetime) and start_time is not None:
        start_time = date_to_datetime(start_time)
    if not isinstance(end_time, dt.datetime) and end_time is not None:
        end_time = date_to_datetime(end_time, dt.time(hour=23, minute=59, second=59))

    if start_time and end_time and start_time > end_time:
        raise ValueError('Start of time interval is larger than end of time interval')

    return start_time, end_time


def serialize_time(timestamp_input, *, use_tz=False):
    """ Transforms datetime objects into ISO 8601 strings

    :param timestamp_input: A datetime object or a tuple of datetime objects
    :type timestamp_input: datetime.date or datetime.datetime or tuple(datetime.date or datetime.datetime)
    :param use_tz: If `True` it will ensure that the serialized string contains a timezone information (typically
        with `Z` at the end instead of +00:00). If `False` it will make sure to remove any timezone information
    :type use_tz: bool
    :return: Timestamp(s) serialized into string(s)
    :rtype: str or tuple(str)
    """
    if isinstance(timestamp_input, tuple):
        return tuple(serialize_time(timestamp, use_tz=use_tz) for timestamp in timestamp_input)

    if timestamp_input is None:
        return '..'

    if not isinstance(timestamp_input, dt.date):
        raise ValueError('Expected a datetime object or a tuple of datetime objects')

    if use_tz and not isinstance(timestamp_input, dt.datetime):
        raise ValueError('Cannot ensure timezone information for datetime.date objects, use datetime.datetime instead')

    if use_tz and not timestamp_input.tzinfo:
        timestamp_input = timestamp_input.replace(tzinfo=dateutil.tz.tzutc())

    if not use_tz and isinstance(timestamp_input, dt.datetime) and timestamp_input.tzinfo:
        timestamp_input = timestamp_input.replace(tzinfo=None)

    return timestamp_input.isoformat().replace('+00:00', 'Z')


def date_to_datetime(date, time=None):
    """ Converts a date object into datetime object

    :param date: a date object
    :type date: datetime.date
    :param time: an option time object, if not provided it will replace it with 00:00:00
    :type time: datetime.time
    :return: A datetime object derived from date and time
    :rtype: datetime.datetime
    """
    if time is None:
        time = dt.datetime.min.time()
    return dt.datetime.combine(date, time)


def filter_times(timestamps, time_difference):
    """ Filters out timestamps within time_difference, preserving only the oldest timestamp.

    :param timestamps: A list of timestamps
    :type timestamps: list(datetime.datetime)
    :param time_difference: A time difference threshold
    :type time_difference: datetime.timedelta
    :return: An ordered list of timestamps `d_1<=d_2<=...<=d_n` such that `d_(i+1)-d_i > time_difference`
    :rtype: list(datetime.datetime)
    """
    timestamps = sorted(set(timestamps))

    filtered_timestamps = []
    for current_timestamp in timestamps:
        if not filtered_timestamps or current_timestamp - filtered_timestamps[-1] > time_difference:
            filtered_timestamps.append(current_timestamp)

    return filtered_timestamps



"""
This module lists all externally useful classes and functions
"""

from .data_request import WmsRequest, WcsRequest, FisRequest, GeopediaWmsRequest, GeopediaImageRequest, \
    AwsTileRequest, AwsProductRequest, get_safe_format, download_safe_format

from .aws import AwsProduct, AwsTile
from .aws_safe import SafeProduct, SafeTile

from .areas import BBoxSplitter, OsmSplitter, TileSplitter, CustomGridSplitter, UtmGridSplitter, UtmZoneSplitter, \
    BatchSplitter

from .ogc import WebFeatureService
from .geopedia import GeopediaFeatureIterator, GeopediaSession

from .geometry import BBox, Geometry, BBoxCollection
from .constants import CustomUrlParam, CRS, MimeType, SHConstants, AwsConstants, ServiceType, \
    HistogramType
from .data_collections import DataCollection, DataSource

from .config import SHConfig

from .download import DownloadRequest, get_json, get_xml, DownloadClient, AwsDownloadClient, \
    SentinelHubDownloadClient, SentinelHubStatisticalDownloadClient

from .exceptions import DownloadFailedException, AwsDownloadFailedException

from .opensearch import get_tile_info_id, get_tile_info, get_area_dates, get_area_info

from .io_utils import read_data, write_data
from .os_utils import get_content_list, get_folder_list, get_file_list, make_folder, create_parent_folder, rename, size
from .geo_utils import bbox_to_dimensions, bbox_to_resolution, get_image_dimension, to_utm_bbox, get_utm_bbox,\
    wgs84_to_utm, to_wgs84, utm_to_pixel, pixel_to_utm, wgs84_to_pixel, get_utm_crs, transform_point
from .time_utils import next_date, prev_date, get_current_date

from .testing_utils import TestSentinelHub, TestCaseContainer

from .sentinelhub_session import SentinelHubSession
from .sentinelhub_batch import SentinelHubBatch
from .sentinelhub_request import SentinelHubRequest
from .sentinelhub_byoc import SentinelHubBYOC, ByocCollection, ByocCollectionAdditionalData, ByocTile
from .sentinelhub_catalog import SentinelHubCatalog
from .sentinelhub_statistical import SentinelHubStatistical

from .time_utils import parse_time, parse_time_interval, serialize_time, is_valid_time, filter_times

from ._version import __version__



"""
Module for working with Sentinel Hub OGC services
"""
import logging
import datetime
from base64 import b64encode
from urllib.parse import urlencode

import shapely.geometry

from .constants import ServiceType, MimeType, CRS, SHConstants, CustomUrlParam
from .config import SHConfig
from .data_collections import DataCollection, handle_deprecated_data_source
from .geo_utils import get_image_dimension
from .geometry import BBox, Geometry
from .download import DownloadRequest, SentinelHubDownloadClient
from .sh_utils import FeatureIterator
from .time_utils import parse_time, parse_time_interval, serialize_time, filter_times

LOGGER = logging.getLogger(__name__)


class OgcImageService:
    """Sentinel Hub OGC services class for providing image data

    Intermediate layer between QGC-type requests (WmsRequest and WcsRequest) and the Sentinel Hub OGC (WMS and WCS)
    services.
    """
    def __init__(self, config=None):
        """
        :param config: A custom instance of config class to override parameters from the saved configuration.
        :type config: SHConfig or None
        """
        self.config = config or SHConfig()
        self.config.raise_for_missing_instance_id()

        self._base_url = self.config.get_sh_ogc_url()
        self.wfs_iterator = None

    def get_request(self, request):
        """ Get download requests

        Create a list of DownloadRequests for all Sentinel-2 acquisitions within request's time interval and
        acceptable cloud coverage.

        :param request: QGC-type request with specified bounding box, time interval, and cloud coverage for specific
                        product.
        :type request: OgcRequest or GeopediaRequest
        :return: list of DownloadRequests
        """
        size_x, size_y = self.get_image_dimensions(request)
        return [DownloadRequest(url=self.get_url(request=request, date=date, size_x=size_x, size_y=size_y),
                                data_type=request.image_format, headers=SHConstants.HEADERS)
                for date in self.get_dates(request)]

    def get_url(self, request, *, date=None, size_x=None, size_y=None, geometry=None):
        """ Returns url to Sentinel Hub's OGC service for the product specified by the OgcRequest and date.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :type request: OgcRequest or GeopediaRequest
        :param date: acquisition date or None
        :type date: datetime.datetime or None
        :param size_x: horizontal image dimension
        :type size_x: int or str
        :param size_y: vertical image dimension
        :type size_y: int or str
        :type geometry: list of BBox or Geometry
        :return:  dictionary with parameters
        :return: url to Sentinel Hub's OGC service for this product.
        :rtype: str
        """
        url = self.get_base_url(request)
        authority = request.theme if hasattr(request, 'theme') else self.config.instance_id

        params = self._get_common_url_parameters(request)
        if request.service_type in (ServiceType.WMS, ServiceType.WCS):
            params = {**params, **self._get_wms_wcs_url_parameters(request, date)}
        if request.service_type is ServiceType.WMS:
            params = {**params, **self._get_wms_url_parameters(request, size_x, size_y)}
        elif request.service_type is ServiceType.WCS:
            params = {**params, **self._get_wcs_url_parameters(request, size_x, size_y)}
        elif request.service_type is ServiceType.FIS:
            params = {**params, **self._get_fis_parameters(request, geometry)}

        return '{}/{}?{}'.format(url, authority, urlencode(params))

    def get_base_url(self, request):
        """ Creates base url string.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :type request: OgcRequest or GeopediaRequest
        :return: base string for url to Sentinel Hub's OGC service for this product.
        :rtype: str
        """
        url = '{}/{}'.format(self._base_url, request.service_type.value)

        if hasattr(request, 'data_collection') and request.data_collection.service_url:
            url = url.replace(self.config.sh_base_url, request.data_collection.service_url)

        return url

    @staticmethod
    def _get_common_url_parameters(request):
        """ Returns parameters common dictionary for WMS, WCS and FIS request.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :type request: OgcRequest or GeopediaRequest
        :return:  dictionary with parameters
        :rtype: dict
        """
        params = {
            'SERVICE': request.service_type.value,
            'WARNINGS': False
        }

        if hasattr(request, 'maxcc'):
            params['MAXCC'] = 100.0 * request.maxcc

        if hasattr(request, 'custom_url_params') and request.custom_url_params is not None:
            params = {**params,
                      **{k.value: str(v) for k, v in request.custom_url_params.items()}}

            if CustomUrlParam.EVALSCRIPT.value in params:
                evalscript = params[CustomUrlParam.EVALSCRIPT.value]
                params[CustomUrlParam.EVALSCRIPT.value] = b64encode(evalscript.encode()).decode()

            if CustomUrlParam.GEOMETRY.value in params:
                geometry = params[CustomUrlParam.GEOMETRY.value]
                crs = request.bbox.crs

                if isinstance(geometry, Geometry):
                    if geometry.crs is not crs:
                        raise ValueError('Geometry object in custom_url_params should have the same CRS as given BBox')
                else:
                    geometry = Geometry(geometry, crs)

                if geometry.crs is CRS.WGS84:
                    geometry = geometry.reverse()

                params[CustomUrlParam.GEOMETRY.value] = geometry.wkt

        return params

    @staticmethod
    def _get_wms_wcs_url_parameters(request, date):
        """ Returns parameters common dictionary for WMS and WCS request.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :type request: OgcRequest or GeopediaRequest
        :param date: acquisition date or None
        :type date: datetime.datetime or None
        :return:  dictionary with parameters
        :rtype: dict
        """
        params = {
            'BBOX': str(request.bbox.reverse()) if request.bbox.crs is CRS.WGS84 else str(request.bbox),
            'FORMAT': MimeType.get_string(request.image_format),
            'CRS': CRS.ogc_string(request.bbox.crs),
        }

        if date is not None:
            start_date = date if request.time_difference < datetime.timedelta(
                seconds=0) else date - request.time_difference
            end_date = date if request.time_difference < datetime.timedelta(
                seconds=0) else date + request.time_difference

            start_date, end_date = serialize_time((start_date, end_date), use_tz=True)
            params['TIME'] = f'{start_date}/{end_date}'

        return params

    @staticmethod
    def _get_wms_url_parameters(request, size_x, size_y):
        """ Returns parameters dictionary for WMS request.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :type request: OgcRequest or GeopediaRequest
        :param size_x: horizontal image dimension
        :type size_x: int or str
        :param size_y: vertical image dimension
        :type size_y: int or str
        :return:  dictionary with parameters
        :rtype: dict
        """
        return {
            'WIDTH': size_x,
            'HEIGHT': size_y,
            'LAYERS': request.layer,
            'REQUEST': 'GetMap',
            'VERSION': '1.3.0'
        }

    @staticmethod
    def _get_wcs_url_parameters(request, size_x, size_y):
        """ Returns parameters dictionary for WCS request.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :type request: OgcRequest or GeopediaRequest
        :param size_x: horizontal image dimension
        :type size_x: int or str
        :param size_y: vertical image dimension
        :type size_y: int or str
        :return:  dictionary with parameters
        :rtype: dict
        """
        return {
            'RESX': size_x,
            'RESY': size_y,
            'COVERAGE': request.layer,
            'REQUEST': 'GetCoverage',
            'VERSION': '1.1.2'
        }

    @staticmethod
    def _get_fis_parameters(request, geometry):
        """ Returns parameters dictionary for FIS request.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :type request: OgcRequest or GeopediaRequest
        :param geometry: list of bounding boxes or geometries
        :type geometry: list of BBox or Geometry
        :return:  dictionary with parameters
        :rtype: dict
        """
        start_time, end_time = serialize_time(parse_time_interval(request.time), use_tz=True)

        params = {
            'CRS': CRS.ogc_string(geometry.crs),
            'LAYER': request.layer,
            'RESOLUTION': request.resolution,
            'TIME': f'{start_time}/{end_time}'
        }

        if not isinstance(geometry, (BBox, Geometry)):
            raise ValueError('Each geometry must be an instance of sentinelhub.{} or sentinelhub.{} but {} '
                             'found'.format(BBox.__name__, Geometry.__name__, geometry))
        if geometry.crs is CRS.WGS84:
            geometry = geometry.reverse()
        if isinstance(geometry, Geometry):
            params['GEOMETRY'] = geometry.wkt
        else:
            params['BBOX'] = str(geometry)

        if request.bins:
            params['BINS'] = request.bins

        if request.histogram_type:
            params['TYPE'] = request.histogram_type.value

        return params

    def get_dates(self, request):
        """ Get available Sentinel-2 acquisitions at least time_difference apart

        List of all available Sentinel-2 acquisitions for given bbox with max cloud coverage and the specified
        time interval. When a single time is specified the request will return that specific date, if it exists.
        If a time range is specified the result is a list of all scenes between the specified dates conforming to
        the cloud coverage criteria. Most recent acquisition being first in the list.

        When a time_difference threshold is set to a positive value, the function filters out all datetimes which
        are within the time difference. The oldest datetime is preserved, all others all deleted.

        :param request: OGC-type request
        :type request: WmsRequest or WcsRequest
        :return: List of dates of existing acquisitions for the given request
        :rtype: list(datetime.datetime) or [None]
        """
        if request.data_collection.is_timeless:
            return [None]

        if request.wfs_iterator is None:
            self.wfs_iterator = WebFeatureService(request.bbox, request.time, data_collection=request.data_collection,
                                                  maxcc=request.maxcc, config=self.config)
        else:
            self.wfs_iterator = request.wfs_iterator

        dates = self.wfs_iterator.get_dates()
        dates = filter_times(dates, request.time_difference)

        LOGGER.debug('Initializing requests for dates: %s', dates)
        return dates

    @staticmethod
    def get_image_dimensions(request):
        """ Verifies or calculates image dimensions.

        :param request: OGC-type request
        :type request: WmsRequest or WcsRequest
        :return: horizontal and vertical dimensions of requested image
        :rtype: (int or str, int or str)
        """
        if request.service_type is ServiceType.WCS or (isinstance(request.size_x, int) and
                                                       isinstance(request.size_y, int)):
            return request.size_x, request.size_y
        if not isinstance(request.size_x, int) and not isinstance(request.size_y, int):
            raise ValueError("At least one of parameters 'width' and 'height' must have an integer value")
        missing_dimension = get_image_dimension(request.bbox, width=request.size_x, height=request.size_y)
        if request.size_x is None:
            return missing_dimension, request.size_y
        if request.size_y is None:
            return request.size_x, missing_dimension
        raise ValueError("Parameters 'width' and 'height' must be integers or None")

    def get_wfs_iterator(self):
        """ Returns iterator over info about all satellite tiles used for the request

        :return: Iterator of dictionaries containing info about all satellite tiles used in the request. In case of
                 DataCollection.DEM it returns None.
        :rtype: Iterator[dict] or None
        """
        return self.wfs_iterator


class WebFeatureService(FeatureIterator):
    """ Class for interaction with Sentinel Hub WFS service

    The class is an iterator over info data of all available satellite tiles for requested parameters. It collects data
    from Sentinel Hub service only during the first iteration. During next iterations it returns already obtained data.
    The data is in the order returned by Sentinel Hub WFS service.
    """
    def __init__(self, bbox, time_interval, *, data_collection=None, maxcc=1.0, data_source=None, config=None):
        """
        :param bbox: Bounding box of the requested image. Coordinates must be in the specified coordinate reference
            system.
        :type bbox: geometry.BBox
        :param time_interval: interval with start and end date of the form YYYY-MM-DDThh:mm:ss or YYYY-MM-DD
        :type time_interval: (str, str)
        :param data_collection: A collection of requested satellite data
        :type data_collection: DataCollection
        :param maxcc: Maximum accepted cloud coverage of an image. Float between 0.0 and 1.0. Default is 1.0.
        :type maxcc: float
        :param config: A custom instance of config class to override parameters from the saved configuration.
        :type config: SHConfig or None
        :param data_source: A deprecated alternative to data_collection
        :type data_source: DataCollection
        """
        self.config = config or SHConfig()
        self.config.raise_for_missing_instance_id()

        self.bbox = bbox

        self.latest_time_only = time_interval == SHConstants.LATEST
        if self.latest_time_only:
            self.time_interval = datetime.datetime(year=1985, month=1, day=1), datetime.datetime.now()
        else:
            self.time_interval = parse_time_interval(time_interval)

        self.data_collection = DataCollection(handle_deprecated_data_source(data_collection, data_source,
                                                                            default=DataCollection.SENTINEL2_L1C))
        self.maxcc = maxcc
        self.max_features_per_request = 1 if self.latest_time_only else self.config.max_wfs_records_per_query

        client = SentinelHubDownloadClient(config=self.config)
        url = self._build_service_url()
        params = self._build_request_params()

        super().__init__(client, url, params)
        self.next = 0

    def _build_service_url(self):
        """ Creates an URL to WFS service
        """
        base_url = self.config.get_sh_ogc_url()

        if self.data_collection.service_url:
            base_url = base_url.replace(self.config.sh_base_url, self.data_collection.service_url)

        return f'{base_url}/{ServiceType.WFS.value}/{self.config.instance_id}'

    def _build_request_params(self):
        """ Builds URL parameters for WFS service
        """
        start_time, end_time = serialize_time(self.time_interval, use_tz=True)
        return {
            'SERVICE': ServiceType.WFS.value,
            'WARNINGS': False,
            'REQUEST': 'GetFeature',
            'TYPENAMES': self.data_collection.wfs_id,
            'BBOX': str(self.bbox.reverse()) if self.bbox.crs is CRS.WGS84 else str(self.bbox),
            'OUTPUTFORMAT': MimeType.JSON.get_string(),
            'SRSNAME': self.bbox.crs.ogc_string(),
            'TIME': f'{start_time}/{end_time}',
            'MAXCC': 100.0 * self.maxcc,
            'MAXFEATURES': self.max_features_per_request
        }

    def _fetch_features(self):
        """ Collects data from WFS service
        """
        params = {
            **self.params,
            'FEATURE_OFFSET': self.next
        }
        url = f'{self.url}?{urlencode(params)}'

        LOGGER.debug("URL=%s", url)
        response = self.client.get_json(url)

        new_features = response['features']

        if len(new_features) < self.max_features_per_request or self.latest_time_only:
            self.finished = True
        else:
            self.next += self.max_features_per_request

        is_sentinel1 = self.data_collection.is_sentinel1
        new_features = [feature_info for feature_info in new_features
                        if not is_sentinel1 or self._sentinel1_product_check(feature_info)]
        return new_features

    def get_dates(self):
        """ Returns a list of acquisition times from tile info data

        :return: List of acquisition times in the order returned by WFS service.
        :rtype: list(datetime.datetime)
        """
        tile_dates = []

        for tile_info in self:
            if not tile_info['properties']['date']:  # could be True for custom (BYOC) data collections
                tile_dates.append(None)
            else:
                date_str = tile_info['properties']['date']
                time_str = tile_info['properties']['time']
                tile_dates.append(parse_time(f'{date_str}T{time_str}'))

        return tile_dates

    def get_geometries(self):
        """ Returns a list of geometries from tile info data

        :return: List of multipolygon geometries in the order returned by WFS service.
        :rtype: list(shapely.geometry.MultiPolygon)
        """
        return [shapely.geometry.shape(tile_info['geometry']) for tile_info in self]

    def get_tiles(self):
        """ Returns list of tiles with tile name, date and AWS index

        :return: List of tiles in form of (tile_name, date, aws_index)
        :rtype: list((str, str, int))
        """
        return [self._parse_tile_url(tile_info['properties']['path']) for tile_info in self]

    @staticmethod
    def _parse_tile_url(tile_url):
        """ Extracts tile name, data and AWS index from tile URL

        :param tile_url: Location of tile at AWS
        :type tile_url: str
        :return: Tuple in a form (tile_name, date, aws_index)
        :rtype: (str, str, int)
        """
        props = tile_url.rsplit('/', 7)
        return ''.join(props[1:4]), '-'.join(props[4:7]), int(props[7])

    def _sentinel1_product_check(self, tile_info):
        """ Checks if Sentinel-1 tile info match the data collection definition
        """
        product_id = tile_info['properties']['id']
        props = product_id.split('_')
        swath_mode, resolution, polarization = props[1], props[2][3], props[3][2:4]
        orbit_direction = tile_info['properties'].get('orbitDirection', '')

        if not (swath_mode in ['IW', 'EW'] and resolution in ['M', 'H'] and polarization in ['DV', 'DH', 'SV', 'SH']):
            raise ValueError(f'Unknown Sentinel-1 tile type: {product_id}')

        return (swath_mode == self.data_collection.swath_mode or self.data_collection.swath_mode is None) \
            and (polarization == self.data_collection.polarization or self.data_collection.polarization is None) \
            and (resolution == self.data_collection.resolution[0] or self.data_collection.resolution is None) \
            and self.data_collection.contains_orbit_direction(orbit_direction)



"""
Module implementing some utility functions not suitable for other utility modules
"""
from abc import ABC, abstractmethod
from urllib.parse import urlencode

from .exceptions import MissingDataInRequestException


class FeatureIterator(ABC):
    """ An implementation of a base feature iteration class

    Main functionalities:

    - The iterator will load only as many features as needed at any moment
    - It will keep downloaded features in memory so that iterating over it again will not have to download the same
      features again.
    """
    def __init__(self, client, url, params=None):
        """
        :param client: An instance of a download client object
        :type client: DownloadClient
        :param url: An URL where requests will be made
        :type url: str
        :param params: Parameters to be sent with each request
        :type params: dict or None
        """
        self.client = client
        self.url = url
        self.params = params or {}

        self.index = 0
        self.features = []
        self.next = None
        self.finished = False

    def __iter__(self):
        """ Method called at the beginning of a new iteration

        :return: It returns the iterator class itself
        :rtype: FeatureIterator
        """
        self.index = 0
        return self

    def __next__(self):
        """ Method called to provide the next feature in iteration

        :return: the next feature
        """
        while self.index >= len(self.features) and not self.finished:
            new_features = self._fetch_features()
            self.features.extend(new_features)

        if self.index < len(self.features):
            self.index += 1
            return self.features[self.index - 1]

        raise StopIteration

    @abstractmethod
    def _fetch_features(self):
        """ Collects and returns more features from the service
        """
        raise NotImplementedError


class SentinelHubFeatureIterator(FeatureIterator):
    """ Feature iterator for the most common implementation of feature pagination at Sentinel Hub services
    """
    def __init__(self, *args, exception_message=None, **kwargs):
        """
        :param args: Arguments passed to FeatureIterator
        :param exception_message: A message to be raise if no feature are found
        :type exception_message: str
        :param kwargs: Keyword arguments passed to FeatureIterator
        """
        self.exception_message = exception_message or 'No data found'

        super().__init__(*args, **kwargs)

    def _fetch_features(self):
        """ Collect more results from the service
        """
        params = remove_undefined({
            **self.params,
            'viewtoken': self.next
        })
        url = f'{self.url}?{urlencode(params)}'

        results = self.client.get_json(url, use_session=True)

        new_features = results.get('data')
        if new_features is None:
            raise MissingDataInRequestException(self.exception_message)

        self.next = results['links'].get('nextToken')
        self.finished = self.next is None or not new_features

        return new_features


def _update_other_args(dict1, dict2):
    """ Function for a recursive update of `dict1` with `dict2`. The function loops over the keys in `dict2` and
    only the non-dict like values are assigned to the specified keys.
    """
    for key, value in dict2.items():
        if isinstance(value, dict) and key in dict1:
            _update_other_args(dict1[key], value)
        else:
            dict1[key] = value


def remove_undefined(payload):
    """ Takes a dictionary and removes keys without value
    """
    return {name: value for name, value in payload.items() if value is not None}



"""
Module for working with Sentinel Hub FIS service
"""
import logging

from .download import DownloadRequest
from .constants import MimeType, SHConstants, RequestType
from .ogc import OgcImageService

LOGGER = logging.getLogger(__name__)


class FisService(OgcImageService):
    """ Sentinel Hub OGC services class for providing FIS data

    Intermediate layer between FIS requests and the Sentinel Hub FIS services.
    """
    def get_request(self, request):
        """ Get download requests

        Create a list of DownloadRequests for all Sentinel-2 acquisitions within request's time interval and
        acceptable cloud coverage.

        :param request: OGC-type request with specified bounding box, time interval, and cloud coverage for specific
            product.
        :type request: OgcRequest or GeopediaRequest
        :return: list of DownloadRequests
        """

        return [self._create_request(request=request, geometry=geometry) for geometry in request.geometry_list]

    def _create_request(self, request, geometry):

        url = self.get_base_url(request)

        headers = {'Content-Type': MimeType.JSON.get_string(), **SHConstants.HEADERS}

        post_data = {**self._get_common_url_parameters(request), **self._get_fis_parameters(request, geometry)}
        post_data = {k.lower(): v for k, v in post_data.items()}  # lowercase required on SH service

        return DownloadRequest(url='{}/{}'.format(url, self.config.instance_id),
                               post_values=post_data,
                               data_type=MimeType.JSON, headers=headers,
                               request_type=RequestType.POST)



"""
Module implementing download client that is adjusted to download from AWS
"""
import logging
import warnings

import boto3
from botocore.exceptions import NoCredentialsError

from ..exceptions import AwsDownloadFailedException
from .client import DownloadClient
from .handlers import fail_missing_file


LOGGER = logging.getLogger(__name__)


class AwsDownloadClient(DownloadClient):
    """ An AWS download client class
    """

    GLOBAL_S3_CLIENT = None

    @fail_missing_file
    def _execute_download(self, request):
        """ Executes a download procedure
        """
        if not self.is_s3_request(request):
            return super()._execute_download(request)

        s3_client = self._get_s3_client()

        response_content = self._do_download(request, s3_client)

        LOGGER.debug('Successful download from %s', request.url)
        return response_content

    def _get_s3_client(self):
        """ Provides a s3 client object
        """
        key_args = {}
        if self.config.aws_access_key_id and self.config.aws_secret_access_key:
            key_args = {
                'aws_access_key_id': self.config.aws_access_key_id,
                'aws_secret_access_key': self.config.aws_secret_access_key
            }

        warnings.filterwarnings('ignore', category=ResourceWarning, message='unclosed.*<ssl.SSLSocket.*>')
        try:
            s3_client = boto3.Session().client('s3', **key_args)
            AwsDownloadClient.GLOBAL_S3_CLIENT = s3_client
        except KeyError as exception:  # Sometimes creation of client fails and we use the global client if it exists
            if AwsDownloadClient.GLOBAL_S3_CLIENT is None:
                raise ValueError('Failed to create a client for download from AWS') from exception
            s3_client = AwsDownloadClient.GLOBAL_S3_CLIENT

        return s3_client

    @staticmethod
    def _do_download(request, s3_client):
        """ Does the download from s3
        """
        _, _, bucket_name, url_key = request.url.split('/', 3)

        try:
            response = s3_client.get_object(Bucket=bucket_name, Key=url_key, RequestPayer='requester')
            return response['Body'].read()
        except NoCredentialsError as exception:
            raise ValueError(
                'The requested data is in Requester Pays AWS bucket. In order to download the data please set '
                'your access key either in AWS credentials file or in sentinelhub config.json file using '
                'command line:\n'
                '$ sentinelhub.config --aws_access_key_id <your AWS key> --aws_secret_access_key '
                '<your AWS secret key>') from exception
        except s3_client.exceptions.NoSuchKey as exception:
            raise AwsDownloadFailedException('File in location %s is missing' % request.url) from exception
        except s3_client.exceptions.NoSuchBucket as exception:
            raise ValueError('Aws bucket %s does not exist' % bucket_name) from exception

    @staticmethod
    def is_s3_request(request):
        """ Checks if data has to be downloaded from AWS s3 bucket

        :return: `True` if url describes location at AWS s3 bucket and `False` otherwise
        :rtype: bool
        """
        return request.url.startswith('s3://')



"""
A download part of the package
"""

from .request import DownloadRequest
from .client import DownloadClient, get_json, get_xml
from .sentinelhub_client import SentinelHubDownloadClient
from .sentinelhub_statistical_client import SentinelHubStatisticalDownloadClient
from .aws_client import AwsDownloadClient



"""
Module implementing a rate-limited multi-threaded download client for downloading from Sentinel Hub service
"""
import logging
import time
from threading import Lock, currentThread

import requests

from .handlers import fail_user_errors, retry_temporal_errors
from .client import DownloadClient
from ..sentinelhub_session import SentinelHubSession
from ..sentinelhub_rate_limit import SentinelHubRateLimit


LOGGER = logging.getLogger(__name__)


class SentinelHubDownloadClient(DownloadClient):
    """ Download client specifically configured for download from Sentinel Hub service
    """
    _CACHED_SESSIONS = {}

    def __init__(self, *, session=None, **kwargs):
        """
        :param session: An OAuth2 session with Sentinel Hub service
        :type session: SentinelHubSession or None
        :param kwargs: Optional parameters from DownloadClient
        """
        super().__init__(**kwargs)

        if session is not None and not isinstance(session, SentinelHubSession):
            raise ValueError(f'A session parameter has to be an instance of {SentinelHubSession.__name__} or None, but '
                             f'{session} was given')
        self.session = session

        self.rate_limit = SentinelHubRateLimit(num_processes=self.config.number_of_download_processes)
        self.lock = None

    def download(self, *args, **kwargs):
        """ The main download method

        :param args: Passed to DownloadClient.download
        :param kwargs: Passed to DownloadClient.download
        """
        # Because the Lock object cannot be pickled we create it only here and remove it afterward
        self.lock = Lock()
        try:
            return super().download(*args, **kwargs)
        finally:
            self.lock = None

    @retry_temporal_errors
    @fail_user_errors
    def _execute_download(self, request):
        """ Executes the download with a single thread and uses a rate limit object, which is shared between all threads
        """
        thread_name = currentThread().getName()

        while True:
            sleep_time = self._execute_thread_safe(self.rate_limit.register_next)

            if sleep_time == 0:
                response = self._do_download(request)

                self._execute_thread_safe(self.rate_limit.update, response.headers)

                if response.status_code != requests.status_codes.codes.TOO_MANY_REQUESTS:
                    response.raise_for_status()

                    LOGGER.debug('%s: Successful download from %s', thread_name, request.url)
                    return response.content
            else:
                LOGGER.debug('%s: Sleeping for %0.2f', thread_name, sleep_time)
                time.sleep(sleep_time)

    def _execute_thread_safe(self, thread_unsafe_function, *args, **kwargs):
        """ Executes a function inside a thread lock and handles potential errors
        """
        if self.lock is None:
            return thread_unsafe_function(*args, **kwargs)

        # pylint: disable=consider-using-with
        self.lock.acquire()
        try:
            return thread_unsafe_function(*args, **kwargs)
        finally:
            self.lock.release()

    def _do_download(self, request):
        """ Runs the download
        """
        return requests.request(
            request.request_type.value,
            url=request.url,
            json=request.post_values,
            headers=self._prepare_headers(request),
            timeout=self.config.download_timeout_seconds
        )

    def _prepare_headers(self, request):
        """ Prepares final headers by potentially joining them with session headers
        """
        if not request.use_session:
            return request.headers

        session_headers = self._execute_thread_safe(self._get_session_headers)
        return {
            **session_headers,
            **request.headers
        }

    def _get_session_headers(self):
        """ Provides up-to-date session headers

        Note that calling session_headers property triggers update if session has expired therefore this has to be
        called in a thread-safe way
        """
        return self.get_session().session_headers

    def get_session(self):
        """ Provides the session object used by the client

        :return: A Sentinel Hub session object
        :rtype: SentinelHubSession
        """
        if self.session:
            return self.session

        cache_key = self.config.sh_client_id, self.config.sh_client_secret, self.config.get_sh_oauth_url()
        if cache_key in SentinelHubDownloadClient._CACHED_SESSIONS:
            session = SentinelHubDownloadClient._CACHED_SESSIONS[cache_key]
        else:
            session = SentinelHubSession(config=self.config)
            SentinelHubDownloadClient._CACHED_SESSIONS[cache_key] = session

        self.session = session
        return session



"""
Module implementing the main download client class
"""
import concurrent.futures
import logging
import warnings
import os
import sys

import requests

from ..config import SHConfig
from ..constants import RequestType, MimeType
from ..decoding import decode_data as decode_data_function
from ..exceptions import DownloadFailedException, SHRuntimeWarning, SHDeprecationWarning
from ..io_utils import read_data, write_data
from .handlers import fail_user_errors, retry_temporal_errors
from .request import DownloadRequest


LOGGER = logging.getLogger(__name__)


class DownloadClient:
    """ A basic download client object

    It does the following:

    - downloads the data with multiple threads in parallel,
    - handles any exceptions that occur during download,
    - decodes downloaded data,
    - reads and writes locally stored/cached data
    """
    def __init__(self, *, redownload=False, raise_download_errors=True, config=None):
        """
        :param redownload: If `True` the data will always be downloaded again. By default this is set to `False` and
            the data that has already been downloaded and saved to an expected location will be read from the
            location instead of being downloaded again.
        :type redownload: bool
        :param raise_download_errors: If `True` any error in download process will be raised as
            `DownloadFailedException`. If `False` failed downloads will only raise warnings.
        :type raise_download_errors: bool
        :param config: An instance of configuration class
        :type config: SHConfig
        """
        self.redownload = redownload
        self.raise_download_errors = raise_download_errors

        self.config = config or SHConfig()

    def download(self, download_requests, max_threads=None, decode_data=True):
        """ Download one or multiple requests, provided as a request list.

        :param download_requests: A list of requests or a single request to be executed.
        :type download_requests: List[DownloadRequest] or DownloadRequest
        :param max_threads: Maximum number of threads to be used for download in parallel. The default is
            `max_threads=None` which will use the number of processors on the system multiplied by 5.
        :type max_threads: int or None
        :param decode_data: If `True` it will decode data otherwise it will return it in binary format.
        :type decode_data: bool
        :return: A list of results or a single result, depending on input parameter `download_requests`
        :rtype: list(object) or object
        """
        is_single_request = isinstance(download_requests, DownloadRequest)
        if is_single_request:
            download_requests = [download_requests]

        with concurrent.futures.ThreadPoolExecutor(max_workers=max_threads) as executor:
            download_list = [
                executor.submit(self._single_download, request, decode_data) for request in download_requests
            ]

        data_list = []
        for future in download_list:
            try:
                data_list.append(future.result())
            except DownloadFailedException as download_exception:
                if self.raise_download_errors:
                    traceback = sys.exc_info()[2]
                    raise download_exception.with_traceback(traceback)

                warnings.warn(str(download_exception), category=SHRuntimeWarning)
                data_list.append(None)

        if is_single_request:
            return data_list[0]
        return data_list

    def _single_download(self, request, decode_data):
        """ Method for downloading a single request
        """
        request.raise_if_invalid()

        request_path, response_path = request.get_storage_paths()

        if not self._is_download_required(request, response_path):
            if request.return_data:
                return read_data(response_path, data_format=request.data_type if decode_data else MimeType.RAW)
            return None

        response_content = self._execute_download(request)

        if request_path and request.save_response and (self.redownload or not os.path.exists(request_path)):
            request_info = request.get_request_params(include_metadata=True)
            write_data(request_path, request_info, data_format=MimeType.JSON)
            LOGGER.debug('Saved request info to %s', request_path)

        if request.save_response:
            write_data(response_path, response_content, data_format=MimeType.RAW)
            LOGGER.debug('Saved data to %s', response_path)

        if request.return_data:
            if decode_data:
                return decode_data_function(response_content, request.data_type)
            return response_content
        return None

    @retry_temporal_errors
    @fail_user_errors
    def _execute_download(self, request):
        """ A default way of executing a single download request
        """
        response = requests.request(
            request.request_type.value,
            url=request.url,
            json=request.post_values,
            headers=request.headers,
            timeout=self.config.download_timeout_seconds
        )

        response.raise_for_status()
        LOGGER.debug('Successful download from %s', request.url)

        return response.content

    def _is_download_required(self, request, response_path):
        """ Checks if download should actually be done
        """
        return (request.save_response or request.return_data) and \
               (self.redownload or response_path is None or not os.path.exists(response_path))

    def get_json(self, url, post_values=None, headers=None, request_type=None, **kwargs):
        """ Download request as JSON data type

        :param url: An URL from where the data will be downloaded
        :type url: str
        :param post_values: A dictionary of parameters for a POST request
        :type post_values: dict or None
        :param headers: A dictionary of additional request headers
        :type headers: dict
        :param request_type: A type of HTTP request to make. If not specified, then it will be a GET request if
            `post_values=None` and a POST request otherwise
        :type request_type: RequestType or None
        :param kwargs: Any other parameters that are passed to DownloadRequest class
        :return: JSON data parsed into Python objects
        :rtype: dict or list or str or None
        """
        json_headers = headers or {}

        if request_type is None:
            request_type = RequestType.GET if post_values is None else RequestType.POST

        if request_type is RequestType.POST and 'Content-Type' not in json_headers:
            json_headers = {
                'Content-Type': MimeType.JSON.get_string(),
                **json_headers
            }

        request = DownloadRequest(url=url, headers=json_headers, request_type=request_type, post_values=post_values,
                                  data_type=MimeType.JSON, **kwargs)

        return self._single_download(request, decode_data=True)

    def get_xml(self, url, **kwargs):
        """ Download request as XML data type

        :param url: url to Sentinel Hub's services or other sources from where the data is downloaded
        :type url: str
        :param kwargs: Any other parameters that are passed to DownloadRequest class
        :return: request response as XML instance
        :rtype: XML instance or None
        """
        request = DownloadRequest(url=url, data_type=MimeType.XML, **kwargs)
        return self._single_download(request, decode_data=True)


def get_json(url, post_values=None, headers=None, request_type=None, download_client_class=DownloadClient, **kwargs):
    """ Download request as JSON data type

    :param url: An URL from where the data will be downloaded
    :type url: str
    :param post_values: A dictionary of parameters for a POST request
    :type post_values: dict or None
    :param headers: A dictionary of additional request headers
    :type headers: dict
    :param request_type: A type of HTTP request to make. If not specified, then it will be a GET request if
        `post_values=None` and a POST request otherwise
    :type request_type: RequestType or None
    :param download_client_class: A class that implements a download client
    :type download_client_class: object
    :param kwargs: Parameters that are passed to a DownloadRequest class
    :return: JSON data parsed into Python objects
    :rtype: object or None
    """
    warnings.warn('Function get_json is deprecated and will soon be removed, please use DownloadClient.get_json '
                  'instead', category=SHDeprecationWarning)

    return download_client_class().get_json(
        url=url,
        post_values=post_values,
        headers=headers,
        request_type=request_type,
        **kwargs
    )


def get_xml(url, download_client_class=DownloadClient):
    """ Download request as XML data type

    :param url: url to Sentinel Hub's services or other sources from where the data is downloaded
    :type url: str
    :return: request response as XML instance
    :rtype: XML instance or None
    :param download_client_class: A class that implements a download client
    :type download_client_class: object
    :raises: RunTimeError
    """
    warnings.warn('Function get_xml is deprecated and will soon be removed, please use DownloadClient.get_xml '
                  'instead', category=SHDeprecationWarning)

    return download_client_class().get_xml(url)



"""
Implementation of Sentinel Hub Process API interface
"""
from .constants import MimeType
from .download import SentinelHubDownloadClient
from .sentinelhub_base_api import SentinelHubBaseApiRequest
from .sh_utils import _update_other_args


class SentinelHubRequest(SentinelHubBaseApiRequest):
    """ Sentinel Hub Process API interface

    For more information check
    `Statistical API documentation <https://docs.sentinel-hub.com/api/latest/api/process/>`__.
    """
    _SERVICE_ENDPOINT = 'process'

    def __init__(self, evalscript, input_data, responses, bbox=None, geometry=None, size=None, resolution=None,
                 **kwargs):
        """
        For details of certain parameters check the
        `Process API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/process>`_.

        :param evalscript: `Evalscript <https://docs.sentinel-hub.com/api/latest/#/Evalscript/>`__.
        :type evalscript: str
        :param input_data: A list of input dictionary objects as described in the API reference. It can be generated
            with the helper method `SentinelHubRequest.input_data`
        :type input_data: List[dict or InputDataDict]
        :param responses: A list of `output.responses` objects as described in the API reference. It can be generated
            with the helper function `SentinelHubRequest.output_response`
        :type responses: List[dict]
        :param bbox: Bounding box describing the area of interest.
        :type bbox: sentinelhub.BBox
        :param geometry: Geometry describing the area of interest.
        :type geometry: sentinelhub.Geometry
        :param size: Size of the image.
        :type size: Tuple[int, int]
        :param resolution: Resolution of the image. It has to be in units compatible with the given CRS.
        :type resolution: Tuple[float, float]
        :param data_folder: location of the directory where the fetched data will be saved.
        :type data_folder: str
        :param config: A custom instance of config class to override parameters from the saved configuration.
        :type config: SHConfig or None
        """
        if not isinstance(evalscript, str):
            raise ValueError("'evalscript' should be a string")

        parsed_mime_type = MimeType.from_string(responses[0]['format']['type'].split('/')[1])
        self.mime_type = MimeType.TAR if len(responses) > 1 else parsed_mime_type

        self.payload = self.body(
            request_bounds=self.bounds(bbox=bbox, geometry=geometry),
            request_data=input_data,
            request_output=self.output(size=size, resolution=resolution, responses=responses),
            evalscript=evalscript
        )

        super().__init__(SentinelHubDownloadClient, **kwargs)

    @staticmethod
    def body(request_bounds, request_data, evalscript, request_output=None, other_args=None):
        """ Generate the Process API request body

        :param request_bounds: A dictionary as generated by `SentinelHubRequest.bounds` helper method.
        :type request_bounds: dict
        :param request_data: A list of dictionaries as generated by `SentinelHubRequest.input_data` helper method.
        :type request_data: List[dict]
        :param evalscript: `Evalscript <https://docs.sentinel-hub.com/api/latest/#/Evalscript/>`__.
        :type evalscript: str
        :param request_output: A dictionary as generated by `SentinelHubRequest.output` helper method.
        :type request_output: dict
        :param other_args: Additional dictionary of arguments. If provided, the resulting dictionary will get updated
            by it.
        :param other_args: dict
        """
        request_body = {
            'input': {
                'bounds': request_bounds,
                'data': request_data
            },
            'evalscript': evalscript
        }

        if request_output is not None:
            request_body['output'] = request_output

        if other_args:
            _update_other_args(request_body, other_args)

        return request_body

    @staticmethod
    def output_response(identifier, response_format, other_args=None):
        """ Generate an element of `output.responses` as described in the Process API reference.

        :param identifier: Identifier of the output response.
        :type identifier: str
        :param response_format: A mime type of one of 'png', 'json', 'jpeg', 'tiff'.
        :type response_format: str or sentinelhub.MimeType
        :param other_args: Additional dictionary of arguments. If provided, the resulting dictionary will get updated
            by it.
        :param other_args: dict
        """
        output_response = {
            'identifier': identifier,
            'format': {
                'type': MimeType(response_format).get_string()
            }
        }

        if other_args:
            _update_other_args(output_response, other_args)

        return output_response

    @staticmethod
    def output(responses, size=None, resolution=None, other_args=None):
        """ Generate an `output` part of the request as described in the Process API reference

        :param responses: A list of objects in `output.responses` as generated by `SentinelHubRequest.output_response`.
        :type responses: List[dict]
        :param size: Size of the image.
        :type size: Tuple[int, int]
        :param resolution: Resolution of the image. It has to be in units compatible with the given CRS.
        :type resolution: Tuple[float, float]
        :param other_args: Additional dictionary of arguments. If provided, the resulting dictionary will get updated
            by it.
        :param other_args: dict
        """
        if size and resolution:
            raise ValueError('Either size or resolution argument should be given, not both.')

        request_output = {
            'responses': responses
        }

        if size:
            request_output['width'], request_output['height'] = size
        if resolution:
            request_output['resx'], request_output['resy'] = resolution

        if other_args:
            _update_other_args(request_output, other_args)

        return request_output



"""
Implementation of Sentinel Hub Process API interface
"""
from .constants import MimeType, RequestType
from .download import DownloadRequest
from .data_collections import OrbitDirection
from .data_request import DataRequest
from .geometry import Geometry, BBox
from .sh_utils import _update_other_args
from .time_utils import parse_time_interval, serialize_time


class SentinelHubBaseApiRequest(DataRequest):
    """ A base class for Sentinel Hub interfaces
    """
    _SERVICE_ENDPOINT = ''

    def create_request(self):
        """ Prepares a download request
        """
        headers = {'content-type': MimeType.JSON.get_string(), 'accept': self.mime_type.get_string()}

        self.download_list = [DownloadRequest(
            request_type=RequestType.POST,
            url=self._get_request_url(),
            post_values=self.payload,
            data_folder=self.data_folder,
            save_response=bool(self.data_folder),
            data_type=self.mime_type,
            headers=headers,
            use_session=True
        )]

    @staticmethod
    def input_data(data_collection, *, time_interval=None, maxcc=None, mosaicking_order=None, upsampling=None,
                   downsampling=None, other_args=None):
        """ Generate the `input data` part of the request body

        :param data_collection: One of supported Process API data collections.
        :type data_collection: DataCollection
        :param time_interval: A time interval with start and end date of the form YYYY-MM-DDThh:mm:ss or YYYY-MM-DD or
            a datetime object
        :type time_interval: (str, str) or (datetime, datetime)
        :param maxcc: Maximum accepted cloud coverage of an image. Float between 0.0 and 1.0. Default is 1.0.
        :type maxcc: float or None
        :param mosaicking_order: Mosaicking order, which has to be either 'mostRecent', 'leastRecent' or 'leastCC'.
        :type mosaicking_order: str or None
        :param upsampling: A type of upsampling to apply on data
        :type upsampling: str
        :param downsampling: A type of downsampling to apply on data
        :type downsampling: str
        :param other_args: Additional dictionary of arguments. If provided, the resulting dictionary will get updated
            by it.
        :param other_args: dict
        :return: A dictionary-like object that also contains additional attributes
        :rtype: InputDataDict
        """
        input_data_dict = {
            'type': data_collection.api_id,
        }

        data_filters = _get_data_filters(data_collection, time_interval, maxcc, mosaicking_order)
        if data_filters:
            input_data_dict['dataFilter'] = data_filters

        processing_params = _get_processing_params(upsampling, downsampling)
        if processing_params:
            input_data_dict['processing'] = processing_params

        if other_args:
            _update_other_args(input_data_dict, other_args)

        return InputDataDict(input_data_dict, service_url=data_collection.service_url)

    @staticmethod
    def bounds(bbox=None, geometry=None, other_args=None):
        """ Generate a `bound` part of the API request

        :param bbox: Bounding box describing the area of interest.
        :type bbox: sentinelhub.BBox
        :param geometry: Geometry describing the area of interest.
        :type geometry: sentinelhub.Geometry
        :param other_args: Additional dictionary of arguments. If provided, the resulting dictionary will get updated
            by it.
        :param other_args: dict
        """
        if bbox is None and geometry is None:
            raise ValueError("'bbox' and/or 'geometry' have to be provided.")

        if bbox and not isinstance(bbox, BBox):
            raise ValueError("'bbox' should be an instance of sentinelhub.BBox")

        if geometry and not isinstance(geometry, Geometry):
            raise ValueError("'geometry' should be an instance of sentinelhub.Geometry")

        if bbox and geometry and bbox.crs != geometry.crs:
            raise ValueError('bbox and geometry should be in the same CRS')

        crs = bbox.crs if bbox else geometry.crs

        request_bounds = {
            'properties': {
                'crs': crs.opengis_string
            }
        }

        if bbox:
            request_bounds['bbox'] = list(bbox)

        if geometry:
            request_bounds['geometry'] = geometry.geojson

        if other_args:
            _update_other_args(request_bounds, other_args)

        return request_bounds

    def _get_request_url(self):
        """ It decides which service URL to query. Restrictions from data collection definitions overrule the
        settings from config object.
        """
        data_collection_urls = tuple({
            input_data_dict.service_url for input_data_dict in self.payload['input']['data']
            if isinstance(input_data_dict, InputDataDict) and input_data_dict.service_url is not None
        })
        if len(data_collection_urls) > 1:
            raise ValueError(f'Given data collections are restricted to different services: {data_collection_urls}\n'
                             f'Try defining data collections without these restrictions')

        base_url = data_collection_urls[0] if data_collection_urls else self.config.sh_base_url
        return f'{base_url}/api/v1/{self._SERVICE_ENDPOINT}'


class InputDataDict(dict):
    """ An input data dictionary which also holds additional attributes
    """
    def __init__(self, input_data_dict, *, service_url=None):
        """
        :param input_data_dict: A normal dictionary with input parameters
        :type input_data_dict: dict
        :param service_url: A service URL defined by a data collection
        :type service_url: str
        """
        super().__init__(input_data_dict)
        self.service_url = service_url

    def __repr__(self):
        """ Modified dictionary representation that also shows additional attributes
        """
        normal_dict_repr = super().__repr__()
        return f'{self.__class__.__name__}({normal_dict_repr}, service_url={self.service_url})'


def _get_data_filters(data_collection, time_interval, maxcc, mosaicking_order):
    """ Builds a dictionary of data filters for Process API
    """
    data_filter = {}

    if time_interval:
        start_time, end_time = serialize_time(parse_time_interval(time_interval, allow_undefined=True), use_tz=True)
        data_filter['timeRange'] = {'from': start_time, 'to': end_time}

    if maxcc is not None:
        if not isinstance(maxcc, float) and (maxcc < 0 or maxcc > 1):
            raise ValueError('maxcc should be a float on an interval [0, 1]')

        data_filter['maxCloudCoverage'] = int(maxcc * 100)

    if mosaicking_order:
        mosaic_order_params = ['mostRecent', 'leastRecent', 'leastCC']

        if mosaicking_order not in mosaic_order_params:
            raise ValueError(f'{mosaicking_order} is not a valid mosaickingOrder parameter, it should be one '
                             f'of: {mosaic_order_params}')

        data_filter['mosaickingOrder'] = mosaicking_order

    return {
        **data_filter,
        **_get_data_collection_filters(data_collection)
    }


def _get_data_collection_filters(data_collection):
    """ Builds a dictionary of filters for Process API from a data collection definition
    """
    filters = {}

    if data_collection.swath_mode:
        filters['acquisitionMode'] = data_collection.swath_mode.upper()

    if data_collection.polarization:
        filters['polarization'] = data_collection.polarization.upper()

    if data_collection.resolution:
        filters['resolution'] = data_collection.resolution.upper()

    if data_collection.orbit_direction and data_collection.orbit_direction.upper() != OrbitDirection.BOTH:
        filters['orbitDirection'] = data_collection.orbit_direction.upper()

    if data_collection.timeliness:
        filters['timeliness'] = data_collection.timeliness

    if data_collection.dem_instance:
        filters['demInstance'] = data_collection.dem_instance

    return filters


def _get_processing_params(upsampling, downsampling):
    """ Builds a dictionary of processing parameters for Process API
    """
    processing_params = {}

    if upsampling:
        processing_params['upsampling'] = upsampling

    if downsampling:
        processing_params['downsampling'] = downsampling

    return processing_params



"""
Module implementing Sentinel Hub session object
"""
import logging
import time

from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session

from .config import SHConfig
from .download.request import DownloadRequest
from .download.handlers import retry_temporal_errors, fail_user_errors


LOGGER = logging.getLogger(__name__)


class SentinelHubSession:
    """ Sentinel Hub authentication class

    The class will do OAuth2 authentication with Sentinel Hub service and store the token. It will make sure that the
    token is never expired by automatically refreshing it if expiry time is close.
    """
    SECONDS_BEFORE_EXPIRY = 60

    def __init__(self, config=None):
        """
        :param config: An instance of package configuration class
        :type config: SHConfig
        """
        self.config = config or SHConfig()

        if not (self.config.sh_client_id and self.config.sh_client_secret):
            raise ValueError("Configuration parameters 'sh_client_id' and 'sh_client_secret' have to be set in order "
                             "to authenticate with Sentinel Hub service")

        self._token = None
        _ = self.token

    @property
    def token(self):
        """ Always up-to-date session's token

        :return: A token in a form of dictionary of parameters
        :rtype: dict
        """
        if self._token and self._token['expires_at'] > time.time() + self.SECONDS_BEFORE_EXPIRY:
            return self._token

        # A request parameter is created only in order for error handling decorators to work correctly
        request = DownloadRequest(url=self.config.get_sh_oauth_url())
        self._token = self._fetch_token(request)

        return self._token

    @property
    def session_headers(self):
        """ Provides session authorization headers

        :return: A dictionary with authorization headers
        :rtype: dict
        """
        return {
            'Authorization': 'Bearer {}'.format(self.token['access_token'])
        }

    @retry_temporal_errors
    @fail_user_errors
    def _fetch_token(self, request):
        """ Collects a new token from Sentinel Hub service
        """
        oauth_client = BackendApplicationClient(client_id=self.config.sh_client_id)

        LOGGER.debug('Creating a new authentication session with Sentinel Hub service')
        with OAuth2Session(client=oauth_client) as oauth_session:
            return oauth_session.fetch_token(
                token_url=request.url,
                client_id=self.config.sh_client_id,
                client_secret=self.config.sh_client_secret
            )



"""
Module defining constants and enumerate types used in the package
"""
import functools
import itertools as it
import mimetypes
import re
import warnings
from enum import Enum, EnumMeta

import utm
import pyproj
from aenum import extend_enum

from .exceptions import SHDeprecationWarning, SHUserWarning
from ._version import __version__


class PackageProps:
    """ Class for obtaining package properties. Currently it supports obtaining package version."""

    @staticmethod
    def get_version():
        """ Returns package version

        :return: package version
        :rtype: str
        """
        return __version__


class ServiceUrl:
    """ Most commonly used Sentinel Hub service URLs
    """
    MAIN = 'https://services.sentinel-hub.com'
    USWEST = 'https://services-uswest2.sentinel-hub.com'
    CREODIAS = 'https://creodias.sentinel-hub.com'
    EOCLOUD = 'http://services.eocloud.sentinel-hub.com'
    MUNDI = 'https://shservices.mundiwebservices.com'


class ServiceType(Enum):
    """ Enum constant class for type of service

    Supported types are WMS, WCS, WFS, AWS, IMAGE
    """
    WMS = 'wms'
    WCS = 'wcs'
    WFS = 'wfs'
    AWS = 'aws'
    IMAGE = 'image'
    FIS = 'fis'
    PROCESSING_API = 'processing'


class CRSMeta(EnumMeta):
    """ Metaclass used for building CRS Enum class
    """
    _UNSUPPORTED_CRS = pyproj.CRS(4326)

    def __new__(mcs, cls, bases, classdict):
        """ This is executed at the beginning of runtime when CRS class is created
        """
        for direction, direction_value in [('N', '6'), ('S', '7')]:
            for zone in range(1, 61):
                classdict[f'UTM_{zone}{direction}'] = f'32{direction_value}{zone:02}'

        return super().__new__(mcs, cls, bases, classdict)

    def __call__(cls, crs_value, *args, **kwargs):
        """ This is executed whenever CRS('something') is called
        """
        # pylint: disable=signature-differs
        crs_value = cls._parse_crs(crs_value)

        if isinstance(crs_value, str) and not cls.has_value(crs_value) and crs_value.isdigit() and len(crs_value) >= 4:
            crs_name = f'EPSG_{crs_value}'
            extend_enum(cls, crs_name, crs_value)

        return super().__call__(crs_value, *args, **kwargs)

    @staticmethod
    def _parse_crs(value):
        """ Method for parsing different inputs representing the same CRS enum. Examples:

        - 4326
        - 'EPSG:3857'
        - {'init': 32633}
        - geojson['crs']['properties']['name'] string (urn:ogc:def:crs:...)
        - pyproj.CRS(32743)
        """
        if isinstance(value, dict) and 'init' in value:
            value = value['init']
        if isinstance(value, pyproj.CRS):
            if value == CRSMeta._UNSUPPORTED_CRS:
                message = 'sentinelhub-py supports only WGS 84 coordinate reference system with ' \
                          'coordinate order lng-lat. Given pyproj.CRS(4326) has coordinate order lat-lng. Be careful ' \
                          'to use the correct order of coordinates.'
                warnings.warn(message, category=SHUserWarning)

            epsg_code = value.to_epsg()
            if epsg_code is not None:
                return str(epsg_code)

            error_message = f'Failed to determine an EPSG code of the given CRS:\n{repr(value)}'
            maybe_epsg = value.to_epsg(min_confidence=0)
            if maybe_epsg is not None:
                error_message = f'{error_message}\nIt might be EPSG {maybe_epsg} but pyproj is not confident ' \
                                'enough.'
            raise ValueError(error_message)

        if isinstance(value, int):
            return str(value)
        if isinstance(value, str):
            if 'urn:ogc:def:crs' in value.lower():
                crs_template = re.compile(r'urn:ogc:def:crs:.+::(?P<code>.+)', re.IGNORECASE)
                value = crs_template.match(value).group("code")
            if value.upper() == 'CRS84':
                return '4326'
            return value.lower().strip('epsg: ')
        return value


class CRS(Enum, metaclass=CRSMeta):
    """ Coordinate Reference System enumerate class

    Available CRS constants are WGS84, POP_WEB (i.e. Popular Web Mercator) and constants in form UTM_<zone><direction>,
    where zone is an integer from [1, 60] and direction is either N or S (i.e. northern or southern hemisphere)
    """
    WGS84 = '4326'
    POP_WEB = '3857'
    #: UTM enum members are defined in CRSMeta.__new__

    def __str__(self):
        """ Method for casting CRS enum into string
        """
        return self.ogc_string()

    def __repr__(self):
        """ Method for retrieving CRS enum representation
        """
        return f"CRS('{self.value}')"

    @classmethod
    def has_value(cls, value):
        """ Tests whether CRS contains a constant defined with string `value`.

        :param value: The string representation of the enum constant.
        :type value: str
        :return: `True` if there exists a constant with string value `value`, `False` otherwise
        :rtype: bool
        """
        return value in cls._value2member_map_

    @property
    def epsg(self):
        """ EPSG code property

        :return: EPSG code of given CRS
        :rtype: int
        """
        return int(self.value)

    def ogc_string(self):
        """ Returns a string of the form authority:id representing the CRS.

        :param self: An enum constant representing a coordinate reference system.
        :type self: CRS
        :return: A string representation of the CRS.
        :rtype: str
        """
        return f'EPSG:{CRS(self).value}'

    @property
    def opengis_string(self):
        """ Returns an URL to OGC webpage where the CRS is defined

        :return: An URL with CRS definition
        :rtype: str
        """
        return f'http://www.opengis.net/def/crs/EPSG/0/{self.epsg}'

    def is_utm(self):
        """ Checks if crs is one of the 64 possible UTM coordinate reference systems.

        :param self: An enum constant representing a coordinate reference system.
        :type self: CRS
        :return: `True` if crs is UTM and `False` otherwise
        :rtype: bool
        """
        return self.name.startswith('UTM')

    @functools.lru_cache(maxsize=5)
    def projection(self):
        """ Returns a projection in form of pyproj class. For better time performance it will cache results of
        5 most recently used CRS classes.

        :return: pyproj projection class
        :rtype: pyproj.Proj
        """
        return pyproj.Proj(self._get_pyproj_projection_def(), preserve_units=True)

    @functools.lru_cache(maxsize=5)
    def pyproj_crs(self):
        """ Returns a pyproj CRS class. For better time performance it will cache results of
        5 most recently used CRS classes.

        :return: pyproj CRS class
        :rtype: pyproj.CRS
        """
        return pyproj.CRS(self._get_pyproj_projection_def())

    @functools.lru_cache(maxsize=10)
    def get_transform_function(self, other, always_xy=True):
        """ Returns a function for transforming geometrical objects from one CRS to another. The function will support
        transformations between any objects that pyproj supports.
        For better time performance this method will cache results of 10 most recently used pairs of CRS classes.

        :param self: Initial CRS
        :type self: CRS
        :param other: Target CRS
        :type other: CRS
        :param always_xy: Parameter that is passed to `pyproj.Transformer` object and defines axis order for
            transformation. The default value `True` is in most cases the correct one.
        :type always_xy: bool
        :return: A projection function obtained from pyproj package
        :rtype: function
        """
        return pyproj.Transformer.from_proj(self.projection(), other.projection(), always_xy=always_xy).transform

    @staticmethod
    def get_utm_from_wgs84(lng, lat):
        """ Convert from WGS84 to UTM coordinate system

        :param lng: Longitude
        :type lng: float
        :param lat: Latitude
        :type lat: float
        :return: UTM coordinates
        :rtype: tuple
        """
        _, _, zone, _ = utm.from_latlon(lat, lng)
        direction = 'N' if lat >= 0 else 'S'
        return CRS[f'UTM_{zone}{direction}']

    def _get_pyproj_projection_def(self):
        """ Returns a pyproj crs definition

        For WGS 84 it ensures lng-lat order
        """
        return '+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs' if self is CRS.WGS84 else self.ogc_string()


class CustomUrlParam(Enum):
    """ Enum class to represent supported custom url parameters of OGC services

    Supported parameters are `SHOWLOGO`, `EVALSCRIPT`, `EVALSCRIPTURL`, `PREVIEW`, `QUALITY`, `UPSAMPLING`,
    `DOWNSAMPLING`, `GEOMETRY` and `WARNINGS`.

    See http://sentinel-hub.com/develop/documentation/api/custom-url-parameters and
    https://www.sentinel-hub.com/develop/documentation/api/ogc_api/wms-parameters for more information.
    """
    SHOWLOGO = 'ShowLogo'
    EVALSCRIPT = 'EvalScript'
    EVALSCRIPTURL = 'EvalScriptUrl'
    PREVIEW = 'Preview'
    QUALITY = 'Quality'
    UPSAMPLING = 'Upsampling'
    DOWNSAMPLING = 'Downsampling'
    GEOMETRY = 'Geometry'
    MINQA = 'MinQA'

    @classmethod
    def has_value(cls, value):
        """ Tests whether CustomUrlParam contains a constant defined with a string `value`

        :param value: The string representation of the enum constant
        :type value: str
        :return: `True` if there exists a constant with a string value `value`, `False` otherwise
        :rtype: bool
        """
        return any(value.lower() == item.value.lower() for item in cls)

    @staticmethod
    def get_string(param):
        """ Get custom url parameter name as string

        :param param: CustomUrlParam enum constant
        :type param: Enum constant
        :return: String describing the file format
        :rtype: str
        """
        return param.value


class HistogramType(Enum):
    """ Enum class for types of histogram supported by Sentinel Hub FIS service

    Supported histogram types are EQUALFREQUENCY, EQUIDISTANT and STREAMING
    """
    EQUALFREQUENCY = 'equalfrequency'
    EQUIDISTANT = 'equidistant'
    STREAMING = 'streaming'


class MimeType(Enum):
    """ Enum class to represent supported file formats

    Supported file formats are TIFF 8-bit, TIFF 16-bit, TIFF 32-bit float, PNG, JPEG, JPEG2000, JSON, CSV, ZIP, HDF5,
    XML, GML, RAW
    """
    TIFF = 'tiff'
    PNG = 'png'
    JPG = 'jpg'
    JP2 = 'jp2'
    JSON = 'json'
    CSV = 'csv'
    ZIP = 'zip'
    HDF = 'hdf'
    XML = 'xml'
    GML = 'gml'
    TXT = 'txt'
    TAR = 'tar'
    RAW = 'raw'
    SAFE = 'safe'

    @property
    def extension(self):
        """ Returns file extension of the MimeType object

        :returns: A file extension string
        :rtype: str
        """
        return self.value

    @staticmethod
    def from_string(mime_type_str):
        """ Parses mime type from a file extension string

        :param mime_type_str: A file extension string
        :type mime_type_str: str
        :return: A mime type enum
        :rtype: MimeType
        """
        if MimeType.has_value(mime_type_str):
            return MimeType(mime_type_str)

        try:
            return {
                'tif': MimeType.TIFF,
                'jpeg': MimeType.JPG,
                'jpg': MimeType.JPG,
                'hdf5': MimeType.HDF,
                'h5': MimeType.HDF
            }[mime_type_str]
        except KeyError as exception:
            raise ValueError(f'Data format .{mime_type_str} is not supported') from exception

    @staticmethod
    def canonical_extension(fmt_ext):
        """ A deprecated method, use MimeType.from_string().extension instead
        """
        warnings.warn('This method is deprecated and will soon be removed, use MimeType.from_string().extension '
                      'instead', category=SHDeprecationWarning)
        return MimeType.from_string(fmt_ext).extension

    def is_image_format(self):
        """ Checks whether file format is an image format

        Example: ``MimeType.PNG.is_image_format()`` or ``MimeType.is_image_format(MimeType.PNG)``

        :param self: File format
        :type self: MimeType
        :return: `True` if file is in image format, `False` otherwise
        :rtype: bool
        """
        return self in frozenset([MimeType.TIFF, MimeType.PNG, MimeType.JP2, MimeType.JPG])

    def is_api_format(self):
        """ Checks if mime type is supported by Sentinel Hub API

        :return: True if API supports this format and False otherwise
        :rtype: bool
        """
        return self in frozenset([MimeType.JPG, MimeType.PNG, MimeType.TIFF, MimeType.JSON])

    @classmethod
    def has_value(cls, value):
        """ Tests whether MimeType contains a constant defined with string ``value``

        :param value: The string representation of the enum constant
        :type value: str
        :return: `True` if there exists a constant with string value ``value``, `False` otherwise
        :rtype: bool
        """
        return value in cls._value2member_map_

    def get_string(self):
        """ Get file format as string

        :return: String describing the file format
        :rtype: str
        """
        if self is MimeType.TAR:
            return 'application/x-tar'
        if self is MimeType.JSON:
            return 'application/json'
        if self is MimeType.JP2:
            return 'image/jpeg2000'
        if self is MimeType.RAW:
            return self.value
        return mimetypes.types_map['.' + self.value]

    def get_expected_max_value(self):
        """ Returns max value of image `MimeType` format and raises an error if it is not an image format

        :return: A maximum value of specified image format
        :rtype: int or float
        :raises: ValueError
        """
        try:
            return {
                MimeType.TIFF: 65535,
                MimeType.PNG: 255,
                MimeType.JPG: 255,
                MimeType.JP2: 10000
            }[self]
        except KeyError as exception:
            raise ValueError(f'Type {self} is not supported by this method') from exception


class RequestType(Enum):
    """ Enum constant class for GET/POST request type """
    GET = 'GET'
    POST = 'POST'
    DELETE = 'DELETE'
    PUT = 'PUT'
    PATCH = 'PATCH'


class SHConstants:
    """ Initialisation of constants used by OGC request.

        Constants are LATEST
    """
    LATEST = 'latest'
    HEADERS = {'User-Agent': f'sentinelhub-py/v{PackageProps.get_version()}'}


class AwsConstants:
    """ Initialisation of every constant used by AWS classes

    For each supported data collection it contains lists of all possible bands and all possible metadata files:

        - S2_L1C_BANDS and S2_L1C_METAFILES
        - S2_L2A_BANDS and S2_L2A_METAFILES

    It also contains dictionary of all possible files and their formats: AWS_FILES
    """
    # General constants:
    SOURCE_ID_LIST = ['L1C', 'L2A']
    TILE_INFO = 'tileInfo'
    PRODUCT_INFO = 'productInfo'
    METADATA = 'metadata'
    PREVIEW = 'preview'
    PREVIEW_JP2 = 'preview*'
    QI_LIST = ['DEFECT', 'DETFOO', 'NODATA', 'SATURA', 'TECQUA']
    QI_MSK_CLOUD = 'qi/MSK_CLOUDS_B00'
    AUX_DATA = 'AUX_DATA'
    DATASTRIP = 'DATASTRIP'
    GRANULE = 'GRANULE'
    HTML = 'HTML'
    INFO = 'rep_info'
    QI_DATA = 'QI_DATA'
    IMG_DATA = 'IMG_DATA'
    INSPIRE = 'inspire'
    MANIFEST = 'manifest'
    TCI = 'TCI'
    PVI = 'PVI'
    ECMWFT = 'auxiliary/ECMWFT'
    AUX_ECMWFT = 'auxiliary/AUX_ECMWFT'
    DATASTRIP_METADATA = 'datastrip/*/metadata'

    # More constants about L2A
    AOT = 'AOT'
    WVP = 'WVP'
    SCL = 'SCL'
    VIS = 'VIS'
    L2A_MANIFEST = 'L2AManifest'
    REPORT = 'report'
    GIPP = 'auxiliary/GIP_TL'
    FORMAT_CORRECTNESS = 'FORMAT_CORRECTNESS'
    GENERAL_QUALITY = 'GENERAL_QUALITY'
    GEOMETRIC_QUALITY = 'GEOMETRIC_QUALITY'
    RADIOMETRIC_QUALITY = 'RADIOMETRIC_QUALITY'
    SENSOR_QUALITY = 'SENSOR_QUALITY'
    QUALITY_REPORTS = [FORMAT_CORRECTNESS, GENERAL_QUALITY, GEOMETRIC_QUALITY, RADIOMETRIC_QUALITY, SENSOR_QUALITY]
    CLASS_MASKS = ['SNW', 'CLD']
    R10m = 'R10m'
    R20m = 'R20m'
    R60m = 'R60m'
    RESOLUTIONS = [R10m, R20m, R60m]
    S2_L2A_BAND_MAP = {R10m: ['B02', 'B03', 'B04', 'B08', AOT, TCI, WVP],
                       R20m: ['B02', 'B03', 'B04', 'B05', 'B06', 'B07', 'B8A', 'B11', 'B12', AOT, SCL, TCI, VIS, WVP],
                       R60m: ['B01', 'B02', 'B03', 'B04', 'B05', 'B06', 'B07', 'B8A', 'B09', 'B11', 'B12', AOT, SCL,
                              TCI, WVP]}

    # Order of elements in following lists is important
    # Sentinel-2 L1C products:
    S2_L1C_BANDS = ['B01', 'B02', 'B03', 'B04', 'B05', 'B06', 'B07', 'B08', 'B8A', 'B09', 'B10', 'B11', 'B12']
    S2_L1C_METAFILES = [PRODUCT_INFO, TILE_INFO, METADATA, INSPIRE, MANIFEST, DATASTRIP_METADATA] +\
                       [PREVIEW, PREVIEW_JP2, TCI] +\
                       [f'{preview}/{band}' for preview, band in it.zip_longest([], S2_L1C_BANDS, fillvalue=PREVIEW)] +\
                       [QI_MSK_CLOUD] +\
                       [f'qi/MSK_{qi}_{band}' for qi, band in it.product(QI_LIST, S2_L1C_BANDS)] + \
                       [f'qi/{qi_report}' for qi_report in [FORMAT_CORRECTNESS, GENERAL_QUALITY,
                                                            GEOMETRIC_QUALITY, SENSOR_QUALITY]] +\
                       [ECMWFT]

    # Sentinel-2 L2A products:
    S2_L2A_BANDS = [f'{resolution}/{band}' for resolution, band_list in sorted(S2_L2A_BAND_MAP.items())
                    for band in band_list]
    S2_L2A_METAFILES = [PRODUCT_INFO, TILE_INFO, METADATA, INSPIRE, MANIFEST, L2A_MANIFEST, REPORT,
                        DATASTRIP_METADATA] + [f'datastrip/*/qi/{qi_report}' for qi_report in QUALITY_REPORTS] +\
                       [f'qi/{source_id}_PVI' for source_id in SOURCE_ID_LIST] +\
                       [f'qi/{mask}_{res.lstrip("R")}' for mask, res in it.product(CLASS_MASKS, [R20m, R60m])] +\
                       [f'qi/MSK_{qi}_{band}' for qi, band in it.product(QI_LIST, S2_L1C_BANDS)] +\
                       [QI_MSK_CLOUD] +\
                       [f'qi/{qi_report}' for qi_report in QUALITY_REPORTS] +\
                       [ECMWFT, AUX_ECMWFT, GIPP]

    # Product files with formats:
    PRODUCT_FILES = {**{PRODUCT_INFO: MimeType.JSON,
                        METADATA: MimeType.XML,
                        INSPIRE: MimeType.XML,
                        MANIFEST: MimeType.SAFE,
                        L2A_MANIFEST: MimeType.XML,
                        REPORT: MimeType.XML,
                        DATASTRIP_METADATA: MimeType.XML},
                     **{f'datastrip/*/qi/{qi_report}': MimeType.XML for qi_report in QUALITY_REPORTS}}
    # Tile files with formats:
    TILE_FILES = {**{TILE_INFO: MimeType.JSON,
                     PRODUCT_INFO: MimeType.JSON,
                     METADATA: MimeType.XML,
                     PREVIEW: MimeType.JPG,
                     PREVIEW_JP2: MimeType.JP2,
                     TCI: MimeType.JP2,
                     QI_MSK_CLOUD: MimeType.GML,
                     ECMWFT: MimeType.RAW,
                     AUX_ECMWFT: MimeType.RAW,
                     GIPP: MimeType.XML},
                  **{f'qi/{qi_report}': MimeType.XML for qi_report in QUALITY_REPORTS},
                  **{f'{preview}/{band}': MimeType.JP2
                     for preview, band in it.zip_longest([], S2_L1C_BANDS, fillvalue=PREVIEW)},
                  **{f'qi/MSK_{qi}_{band}': MimeType.GML for qi, band in it.product(QI_LIST, S2_L1C_BANDS)},
                  **{band: MimeType.JP2 for band in S2_L1C_BANDS},
                  **{band: MimeType.JP2 for band in S2_L2A_BANDS},
                  **{f'qi/{source_id}_PVI': MimeType.JP2 for source_id in SOURCE_ID_LIST},
                  **{f'qi/{mask}_{res.lstrip("R")}': MimeType.JP2 for mask, res in it.product(CLASS_MASKS,
                                                                                              [R20m, R60m])}}

    # All files joined together
    AWS_FILES = {**PRODUCT_FILES,
                 **{filename.split('/')[-1]: data_format for filename, data_format in PRODUCT_FILES.items()},
                 **TILE_FILES,
                 **{filename.split('/')[-1]: data_format for filename, data_format in TILE_FILES.items()}}


class EsaSafeType(Enum):
    """ Enum constants class for ESA .SAFE type.

     Types are OLD_TYPE and COMPACT_TYPE
    """
    OLD_TYPE = 'old_type'
    COMPACT_TYPE = 'compact_type'



"""
Implementation of Sentinel Hub Statistical API interface
"""
from .constants import MimeType
from .download.sentinelhub_statistical_client import SentinelHubStatisticalDownloadClient
from .sentinelhub_base_api import SentinelHubBaseApiRequest
from .sh_utils import _update_other_args
from .time_utils import parse_time_interval, serialize_time


class SentinelHubStatistical(SentinelHubBaseApiRequest):
    """ Sentinel Hub Statistical API interface

    For more information check
    `Statistical API documentation <https://docs.sentinel-hub.com/api/latest/api/statistical/>`__.
    """
    _SERVICE_ENDPOINT = 'statistics'

    def __init__(self, aggregation, input_data, bbox=None, geometry=None, calculations=None, **kwargs):
        """
        For details of certain parameters check the
        `Statistical API reference <https://docs.sentinel-hub.com/api/latest/reference/#tag/statistical>`_.

        :param aggregation: Aggregation part of the payload, which can be generated with `aggregation` method
        :type aggregation: dict
        :param input_data: A list of input dictionary objects as described in the API reference. It can be generated
            with `input_data` method
        :type input_data: List[dict or InputDataDict]
        :param bbox: A bounding box of the request
        :type bbox: sentinelhub.BBox or None
        :param geometry: A geometry of the request
        :type geometry: sentinelhub.Geometry or None
        :param calculations: Calculations part of the payload.
        :param calculations: dict
        :param data_folder: Location of the directory where the downloaded data could be saved.
        :type data_folder: str
        :param config: A custom instance of config class to override parameters from the saved configuration.
        :type config: SHConfig or None
        """
        self.mime_type = MimeType.JSON

        self.payload = self.body(
            request_bounds=self.bounds(bbox=bbox, geometry=geometry),
            request_data=input_data,
            aggregation=aggregation,
            calculations=calculations
        )

        super().__init__(SentinelHubStatisticalDownloadClient, **kwargs)

    @staticmethod
    def body(request_bounds, request_data, aggregation, calculations, other_args=None):
        """ Generate the Process API request body

        :param request_bounds: A dictionary as generated by `bounds` helper method.
        :type request_bounds: dict
        :param request_data: A list of dictionaries as generated by `input_data` helper method.
        :type request_data: List[dict]
        :param aggregation: A dictionary as generated by `aggregation` helper method.
        :type aggregation: dict
        :param calculations: A dictionary defining calculations part of the payload
        :type calculations: dict
        :param other_args: Additional dictionary of arguments. If provided, the resulting dictionary will get updated
            by it.
        :param other_args: dict
        :returns: Request payload dictionary
        :rtype: dict
        """
        # Some parts of the payload have to be defined:
        for input_data_payload in request_data:
            if 'dataFilter' not in input_data_payload:
                input_data_payload['dataFilter'] = {}

        if calculations is None:
            calculations = {
                'default': {}
            }

        request_body = {
            'input': {
                'bounds': request_bounds,
                'data': request_data
            },
            'aggregation': aggregation,
            'calculations': calculations
        }

        if other_args:
            _update_other_args(request_body, other_args)

        return request_body

    @staticmethod
    def aggregation(evalscript, time_interval, aggregation_interval, size=None, resolution=None, other_args=None):
        """ Generate the `aggregation` part of the Statistical API request body

        :param evalscript: An `evalscript <https://docs.sentinel-hub.com/api/latest/#/Evalscript/>`_.
        :type evalscript: str
        :param time_interval: An interval with start and end date of the form YYYY-MM-DDThh:mm:ss or YYYY-MM-DD or
            a datetime object
        :type time_interval: (str, str) or (datetime, datetime)
        :param aggregation_interval: How data from given time interval is aggregated together
        :type aggregation_interval: str
        :param size: A width and height of an image from which data will be aggregated
        :type size: (int, int) or None
        :param resolution: A resolution in x and y dimensions of an image from which data will be aggregated.
            Resolution has to be defined in the same units as request bbox or geometry
        :type resolution: (float, float) or (int, int) or None
        :param other_args: Additional dictionary of arguments. If provided, the resulting dictionary will get updated
            by it.
        :param other_args: dict
        :returns: Aggregation payload dictionary
        :rtype: dict
        """
        start_time, end_time = serialize_time(parse_time_interval(time_interval, allow_undefined=True), use_tz=True)

        payload = {
            'evalscript': evalscript,
            'timeRange': {
                'from': start_time,
                'to': end_time
            },
            'aggregationInterval': {
                'of': aggregation_interval
            }
        }

        if size:
            payload['width'], payload['height'] = size
        if resolution:
            payload['resx'], payload['resy'] = resolution

        if other_args:
            _update_other_args(payload, other_args)

        return payload



"""
Download process for Sentinel Hub Statistical API
"""
import copy
import os
import concurrent.futures

from .sentinelhub_client import SentinelHubDownloadClient
from ..constants import MimeType
from ..decoding import decode_data as decode_data_function
from ..io_utils import read_data, write_data


class SentinelHubStatisticalDownloadClient(SentinelHubDownloadClient):
    """ A special download client for Sentinel Hub Statistical API

    Besides a normal download from Sentinel Hub services it implements an additional process of retrying and caching
    """
    _RETIRABLE_ERRORS = ['EXECUTION_ERROR', 'TIMEOUT']

    def __init__(self, *args, n_interval_retries=1, max_retry_threads=5, **kwargs):
        """
        :param n_interval_retries: Number of retries if a request fails just for a certain timestamp. (This parameter
            is experimental and might be changed in the future.)
        :type n_interval_retries: int
        :param max_retry_threads: Number of threads used for retrying. (This parameter is experimental and might be
            changed in the future.)
        :type max_retry_threads: int
        """
        super().__init__(*args, **kwargs)

        self.n_interval_retries = n_interval_retries
        self.max_retry_threads = max_retry_threads

    def _single_download(self, request, _):
        """ Method for downloading a single request
        """
        request.raise_if_invalid()
        if not (request.save_response or request.return_data):
            return None

        request_path, response_path = request.get_storage_paths()

        download_required = self.redownload or response_path is None or not os.path.exists(response_path)
        if download_required:
            response_content = self._execute_download(request)
            stats_response = decode_data_function(response_content, request.data_type)
        else:
            stats_response = read_data(response_path, data_format=request.data_type)

        failed_time_intervals = {}
        for index, stat_info in enumerate(stats_response['data']):
            if self._has_retriable_error(stat_info):
                failed_time_intervals[index] = stat_info['interval']

        n_succeeded_intervals = 0
        if failed_time_intervals:
            retried_responses = self._download_per_interval(request, failed_time_intervals)
            n_succeeded_intervals = sum('error' not in stat_info for stat_info in retried_responses.values())

            stats_response['data'] = [
                retried_responses.get(index, stat_info) for index, stat_info in enumerate(stats_response['data'])
            ]

        if request_path and request.save_response and (self.redownload or not os.path.exists(request_path)):
            request_info = request.get_request_params(include_metadata=True)
            write_data(request_path, request_info, data_format=MimeType.JSON)

        if request.save_response and (download_required or n_succeeded_intervals > 0):
            write_data(response_path, stats_response, data_format=request.data_type)

        if request.return_data:
            return stats_response
        return None

    def _download_per_interval(self, request, time_intervals):
        """ Download statistics per each time interval
        """
        interval_requests = []
        for time_interval in time_intervals.values():
            interval_request = copy.deepcopy(request)
            interval_request.post_values['aggregation']['timeRange'] = time_interval
            interval_requests.append(interval_request)

        with concurrent.futures.ThreadPoolExecutor(max_workers=self.max_retry_threads) as executor:
            stat_info_responses = list(executor.map(self._execute_single_stat_download, interval_requests))

        return dict(zip(time_intervals, stat_info_responses))

    def _execute_single_stat_download(self, request):
        """ Makes sure a download for a single time interval is retried
        """
        for retry_count in range(self.n_interval_retries):
            response = self._execute_download(request)
            stat_response = decode_data_function(response, request.data_type)
            stat_info = stat_response['data'][0]

            if not self._has_retriable_error(stat_info) or retry_count == self.n_interval_retries - 1:
                return stat_info

        raise ValueError('No retries done')

    def _has_retriable_error(self, stat_info):
        """ Checks if a dictionary of Stat API info for a single time interval has an error that can fixed by retrying
        a request
        """
        error_type = stat_info.get('error', {}).get('type')
        return error_type in self._RETIRABLE_ERRORS



"""
Module implementing an interface with Sentinel Hub Bring your own COG service
"""
from dataclasses import field, dataclass
from datetime import datetime
from typing import Optional, Union

from dataclasses_json import config as dataclass_config
from dataclasses_json import dataclass_json, LetterCase, Undefined, CatchAll

from .data_collections import DataCollection
from .config import SHConfig
from .constants import RequestType, MimeType
from .download.sentinelhub_client import SentinelHubDownloadClient
from .geometry import Geometry
from .sh_utils import SentinelHubFeatureIterator, remove_undefined
from .time_utils import parse_time, serialize_time


datetime_config = dataclass_config(
    encoder=lambda time: serialize_time(time, use_tz=True) if time else None,
    decoder=lambda time: parse_time(time, force_datetime=True) if time else None,
    letter_case=LetterCase.CAMEL
)

geometry_config = dataclass_config(encoder=Geometry.get_geojson,
                                   decoder=lambda x: Geometry.from_geojson(x) if x else None,
                                   exclude=lambda x: x is None, letter_case=LetterCase.CAMEL)


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.INCLUDE)
@dataclass
class ByocCollectionAdditionalData:
    """ Dataclass to hold BYOC collection additional data
    """
    other_data: CatchAll
    bands: Optional[dict] = None
    max_meters_per_pixel: Optional[float] = None
    max_meters_per_pixel_override: Optional[float] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.INCLUDE)
@dataclass
class ByocCollection:
    """ Dataclass to hold BYOC collection data
    """
    name: str
    s3_bucket: str
    other_data: CatchAll
    collection_id: Optional[str] = field(metadata=dataclass_config(field_name='id'), default=None)
    user_id: Optional[str] = None
    created: Optional[datetime] = field(metadata=datetime_config, default=None)
    additional_data: Optional[ByocCollectionAdditionalData] = None
    no_data: Optional[Union[int, float]] = None

    def to_data_collection(self):
        """ Returns a DataCollection enum for this BYOC collection
        """
        if self.additional_data and self.additional_data.bands:
            band_names = tuple(self.additional_data.bands)
            return DataCollection.define_byoc(collection_id=self.collection_id, bands=band_names)

        return DataCollection.define_byoc(collection_id=self.collection_id)


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.INCLUDE)
@dataclass
class ByocTile:
    """ Dataclass to hold BYOC tile data
    """
    path: str
    other_data: CatchAll
    status: Optional[str] = None
    tile_id: Optional[str] = field(metadata=dataclass_config(field_name='id'), default=None)
    tile_geometry: Optional[Geometry] = field(metadata=geometry_config, default=None)
    cover_geometry: Optional[Geometry] = field(metadata=geometry_config, default=None)
    created: Optional[datetime] = field(metadata=datetime_config, default=None)
    sensing_time: Optional[datetime] = field(metadata=datetime_config, default=None)
    additional_data: Optional[dict] = None


class SentinelHubBYOC:
    """ An interface class for Sentinel Hub Bring your own COG (BYOC) API

    For more info check `BYOC API reference
    <https://docs.sentinel-hub.com/api/latest/reference/#tag/byoc_collection>`__.
    """
    def __init__(self, base_url=None, config=None):
        self.config = config or SHConfig()

        base_url = base_url or self.config.sh_base_url
        self.byoc_url = f'{base_url.rstrip("/")}/api/v1/byoc'

        self.client = SentinelHubDownloadClient(config=self.config)

    def iter_collections(self, search=None, **kwargs):
        """ Retrieve collections

        `BYOC API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/getByocCollections>`__

        :param search: A search query
        :param kwargs: Any other request parameters
        :return: iterator over collections
        """
        return SentinelHubFeatureIterator(
            client=self.client,
            url=f'{self.byoc_url}/collections',
            params={
                'search': search,
                **kwargs
            },
            exception_message='Failed to obtain information about available BYOC collections'
        )

    def get_collection(self, collection):
        """ Get collection by its id

        `BYOC API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/getByocCollectionById>`__

        :param collection: a ByocCollection, dict or collection id string
        :return: dictionary of the collection
        :rtype: dict
        """
        url = f'{self.byoc_url}/collections/{self._parse_id(collection)}'
        return self.client.get_json(url=url, use_session=True)['data']

    def create_collection(self, collection):
        """ Create a new collection

        `BYOC API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/createByocCollection>`__

        :param collection: ByocCollection object or a dictionary
        :return: dictionary of the created collection
        :rtype: dict
        """
        coll = self._to_dict(collection)
        url = f'{self.byoc_url}/collections'
        return self.client.get_json(url=url, post_values=coll, use_session=True)['data']

    def update_collection(self, collection):
        """ Update an existing collection

        `BYOC API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/updateByocCollectionById>`__

        :param collection: ByocCollection object or a dictionary
        """
        coll = self._to_dict(collection)
        url = f'{self.byoc_url}/collections/{self._parse_id(coll)}'
        headers = {'Content-Type': MimeType.JSON.get_string()}
        return self.client.get_json(url=url, request_type=RequestType.PUT, post_values=coll,
                                    headers=headers, use_session=True)

    def delete_collection(self, collection):
        """ Delete existing collection by its id

        `BYOC API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/deleteByocCollectionById>`__

        :param collection: a ByocCollection, dict or collection id string
        """
        url = f'{self.byoc_url}/collections/{self._parse_id(collection)}'
        return self.client.get_json(url=url, request_type=RequestType.DELETE, use_session=True)

    def copy_tiles(self, from_collection, to_collection):
        """ Copy tiles from one collection to another

        `BYOC API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/copyByocCollectionTiles>`__

        :param from_collection: a ByocCollection, dict or collection id string
        :param to_collection: a ByocCollection, dict or collection id string
        """
        url = f'{self.byoc_url}/collections/{self._parse_id(from_collection)}' \
              f'/copyTiles?toCollection={self._parse_id(to_collection)}'
        return self.client.get_json(url=url, request_type=RequestType.POST, use_session=True)

    def iter_tiles(self, collection, sort=None, path=None, **kwargs):
        """ Iterator over collection tiles

        `BYOC API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/getByocCollectionTiles>`__

        :param collection: a ByocCollection, dict or collection id string
        :param sort: Order in which to return tiles
        :param path: An exact path where tiles are located
        :param kwargs: Any other request parameters
        :return: iterator
        """
        collection_id = self._parse_id(collection)
        return SentinelHubFeatureIterator(
            client=self.client,
            url=f'{self.byoc_url}/collections/{collection_id}/tiles',
            params={
                'sort': sort,
                'path': path,
                **kwargs
            },
            exception_message=f'Failed to obtain information about tiles in BYOC collection {collection_id}'
        )

    def get_tile(self, collection, tile):
        """ Get a tile of collection

        `BYOC API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/getByocCollectionTileById>`__

        :param collection: a ByocCollection, dict or collection id string
        :param tile: a ByocTile, dict or tile id string
        :return: dictionary of the tile
        :rtype: dict
        """
        url = f'{self.byoc_url}/collections/{self._parse_id(collection)}/tiles/{self._parse_id(tile)}'
        return self.client.get_json(url=url, use_session=True)['data']

    def create_tile(self, collection, tile):
        """ Create tile within collection

        `BYOC API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/createByocCollectionTile>`__

        :param collection: a ByocCollection, dict or collection id string
        :param tile: a ByocTile or dict
        :return: dictionary of the tile
        :rtype: dict
        """
        _tile = self._to_dict(tile)
        url = f'{self.byoc_url}/collections/{self._parse_id(collection)}/tiles'
        return self.client.get_json(url=url, post_values=_tile, use_session=True)['data']

    def update_tile(self, collection, tile):
        """ Update a tile within collection

        `BYOC API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#operation/updateByocCollectionTileById>`__

        :param collection: a ByocCollection, dict or collection id string
        :param tile: a ByocTile or dict
        """
        url = f'{self.byoc_url}/collections/{self._parse_id(collection)}/tiles/{self._parse_id(tile)}'
        headers = {'Content-Type': MimeType.JSON.get_string()}

        _tile = self._to_dict(tile)
        updates = remove_undefined({
            'path': _tile['path'],
            'coverGeometry': _tile.get('coverGeometry'),
            'sensingTime':  _tile.get('sensingTime')
        })

        return self.client.get_json(url=url, request_type=RequestType.PUT, post_values=updates,
                                    headers=headers, use_session=True)

    def delete_tile(self, collection, tile):
        """ Delete a tile from collection

        `BYOC API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#operation/deleteByocCollectionTileById>`__

        :param collection: a ByocCollection, dict or collection id string
        :param tile: a ByocTile, dict or tile id string
        """
        url = f'{self.byoc_url}/collections/{self._parse_id(collection)}/tiles/{self._parse_id(tile)}'
        return self.client.get_json(url=url, request_type=RequestType.DELETE, use_session=True)

    @staticmethod
    def _parse_id(data):
        if isinstance(data, (ByocCollection, DataCollection)):
            return data.collection_id
        if isinstance(data, ByocTile):
            return data.tile_id
        if isinstance(data, dict):
            return data['id']
        if isinstance(data, str):
            return data
        raise ValueError(f'Expected a BYOC/Data dataclass, dictionary or a string, got {data}.')

    @staticmethod
    def _to_dict(data):
        """ Constructs dict from an object (either dataclass or dict)
        """
        if isinstance(data, (ByocCollection, ByocTile, ByocCollectionAdditionalData)):
            return data.to_dict()
        if isinstance(data, dict):
            return data
        raise ValueError(f'Expected either a data class (e.g., ByocCollection, ByocTile) or a dict, got {data}.')



"""
Module defining data collections
"""
import re
import warnings
from enum import Enum, EnumMeta
from typing import Tuple
from dataclasses import dataclass, asdict, field

from aenum import extend_enum

from .config import SHConfig
from .constants import ServiceUrl
from .exceptions import SHDeprecationWarning


class _CollectionType:
    """ Types of Sentinel Hub data collections
    """
    SENTINEL1 = 'Sentinel-1'
    SENTINEL2 = 'Sentinel-2'
    SENTINEL3 = 'Sentinel-3'
    SENTINEL5P = 'Sentinel-5P'
    LANDSAT15 = 'Landsat 1-5'
    LANDSAT45 = 'Landsat 4-5'
    LANDSAT5 = 'Landsat 5'
    LANDSAT7 = 'Landsat 7'
    LANDSAT8 = 'Landsat 8'
    MODIS = 'MODIS'
    ENVISAT_MERIS = 'Envisat Meris'
    DEM = 'DEM'
    BYOC = 'BYOC'
    BATCH = 'BATCH'


class _SensorType:
    """ Satellite sensors
    """
    # pylint: disable=invalid-name
    MSI = 'MSI'
    OLI_TIRS = 'OLI-TIRS'
    TM = 'TM'
    ETM = 'ETM+'
    MSS = 'MSS'
    C_SAR = 'C-SAR'
    OLCI = 'OLCI'
    SLSTR = 'SLSTR'
    TROPOMI = 'TROPOMI'


class _ProcessingLevel:
    """ Processing levels
    """
    # pylint: disable=invalid-name
    L1 = 'L1'
    L2 = 'L2'
    L1B = 'L1B'
    L1C = 'L1C'
    L2A = 'L2A'
    L3B = 'L3B'
    GRD = 'GRD'


class _SwathMode:
    """ Swath modes for SAR sensors
    """
    # pylint: disable=invalid-name
    IW = 'IW'
    EW = 'EW'
    SM = 'SM'
    WV = 'WV'


class _Polarization:
    """ SAR polarizations
    """
    # pylint: disable=invalid-name
    DV = 'DV'
    DH = 'DH'
    SV = 'SV'
    SH = 'SH'
    HH = 'HH'
    HV = 'HV'
    VV = 'VV'
    VH = 'VH'


class _Resolution:
    """ Product resolution (specific to Sentinel-1 collections)
    """
    MEDIUM = 'MEDIUM'
    HIGH = 'HIGH'


class OrbitDirection:
    """ Orbit directions
    """
    ASCENDING = 'ASCENDING'
    DESCENDING = 'DESCENDING'
    BOTH = 'BOTH'


class _Bands:
    """ Different collections of bands
    """
    SENTINEL1_IW = ('VV', 'VH')
    SENTINEL1_EW = ('HH', 'HV')
    SENTINEL1_EW_SH = ('HH',)
    SENTINEL2_L1C = ('B01', 'B02', 'B03', 'B04', 'B05', 'B06', 'B07', 'B08', 'B8A', 'B09', 'B10', 'B11', 'B12')
    SENTINEL2_L2A = ('B01', 'B02', 'B03', 'B04', 'B05', 'B06', 'B07', 'B08', 'B8A', 'B09', 'B11', 'B12')
    SENTINEL3_OLCI = tuple(f'B{str(index).zfill(2)}' for index in range(1, 22))
    SENTINEL3_SLSTR = ('S1', 'S2', 'S3', 'S4', 'S5', 'S6', 'S7', 'S8', 'S9', 'F1', 'F2')
    SENTINEL5P = ('AER_AI_340_380', 'AER_AI_354_388', 'CLOUD_BASE_HEIGHT', 'CLOUD_BASE_PRESSURE', 'CLOUD_FRACTION',
                  'CLOUD_OPTICAL_THICKNESS', 'CLOUD_TOP_HEIGHT', 'CLOUD_TOP_PRESSURE', 'CO', 'HCHO', 'NO2', 'O3',
                  'SO2', 'CH4')
    LANDSAT15 = ('B01', 'B02', 'B03', 'B04')
    LANDSAT45 = ('B01', 'B02', 'B03', 'B04', 'B05', 'B06', 'B07')
    LANDSAT7_L1 = ('B01', 'B02', 'B03', 'B04', 'B05', 'B06_VCID1', 'B06_VCID2', 'B07', 'B08')
    LANDSAT7_L2 = ('B01', 'B02', 'B03', 'B04', 'B05', 'B06', 'B07')
    LANDSAT8 = ('B01', 'B02', 'B03', 'B04', 'B05', 'B06', 'B07', 'B08', 'B09', 'B10', 'B11')
    LANDSAT8_L2 = ('B01', 'B02', 'B03', 'B04', 'B05', 'B06', 'B07', 'B10')
    DEM = ('DEM',)
    MODIS = ('B01', 'B02', 'B03', 'B04', 'B05', 'B06', 'B07')


class _DataCollectionMeta(EnumMeta):
    """ Meta class that builds DataCollection class enums
    """
    def __call__(cls, value, *args, **kwargs):
        """ This is executed whenever DataCollection('something') is called

        This implements a shortcut to create a BYOC data collection by calling DataCollection('<BYOC collection ID>')
        """
        # This solves a problem of pickling a custom DataCollection and unpickling it in another process
        if isinstance(value, DataCollectionDefinition) and value not in cls._value2member_map_ and value._name:
            cls._try_add_data_collection(value._name, value)

        # pylint: disable=signature-differs
        if not isinstance(value, str):
            return super().__call__(value, *args, **kwargs)

        byoc_data_collection = cls.define_byoc(value)
        message = 'This way of defining a BYOC data collection is deprecated and will soon be removed. Please use ' \
                  'DataCollection.define_byoc(collection_id) instead'
        warnings.warn(message, category=SHDeprecationWarning)

        return byoc_data_collection


@dataclass(frozen=True)
class DataCollectionDefinition:
    """ An immutable definition of a data collection

    Check `DataCollection.define` for more info about attributes of this class
    """
    # pylint: disable=too-many-instance-attributes
    api_id: str = None
    catalog_id: str = None
    wfs_id: str = None
    service_url: str = None
    collection_type: str = None
    sensor_type: str = None
    processing_level: str = None
    swath_mode: str = None
    polarization: str = None
    resolution: str = None
    orbit_direction: str = None
    timeliness: str = None
    bands: Tuple[str, ...] = None
    collection_id: str = None
    is_timeless: bool = False
    has_cloud_coverage: bool = False
    dem_instance: str = None
    # The following parameter is used to preserve custom DataCollection name during pickling and unpickling process:
    _name: str = field(default=None, compare=False)

    def __post_init__(self):
        """ In case a list of bands has been given this makes sure to cast it into a tuple
        """
        if isinstance(self.bands, list):
            object.__setattr__(self, 'bands', tuple(self.bands))

    def __repr__(self):
        """ A nicer representation of parameters that define a data collection
        """
        valid_params = {name: value for name, value in asdict(self).items() if value is not None}
        params_repr = '\n  '.join(f'{name}: {value}' for name, value in valid_params.items() if name != '_name')
        return f'{self.__class__.__name__}(\n  {params_repr}\n)'

    def derive(self, **params):
        """ Create a new data collection definition from current definition and parameters that override current
        parameters

        :param params: Any of DataCollectionDefinition attributes
        :return: A new data collection definition
        :rtype: DataCollectionDefinition
        """
        derived_params = asdict(self)
        for name, value in params.items():
            derived_params[name] = value

        return DataCollectionDefinition(**derived_params)


class DataCollection(Enum, metaclass=_DataCollectionMeta):
    """ An enum class for data collections

    It contains a number of predefined data collections, which are the most commonly used with Sentinel Hub service.
    Additionally it also allows defining new data collections by specifying data collection parameters relevant for
    the service. Check `DataCollection.define` and similar methods for more.
    """
    SENTINEL2_L1C = DataCollectionDefinition(
        api_id='sentinel-2-l1c',
        catalog_id='sentinel-2-l1c',
        wfs_id='DSS1',
        collection_type=_CollectionType.SENTINEL2,
        sensor_type=_SensorType.MSI,
        processing_level=_ProcessingLevel.L1C,
        bands=_Bands.SENTINEL2_L1C,
        has_cloud_coverage=True
    )
    SENTINEL2_L2A = DataCollectionDefinition(
        api_id='sentinel-2-l2a',
        catalog_id='sentinel-2-l2a',
        wfs_id='DSS2',
        collection_type=_CollectionType.SENTINEL2,
        sensor_type=_SensorType.MSI,
        processing_level=_ProcessingLevel.L2A,
        bands=_Bands.SENTINEL2_L2A,
        has_cloud_coverage=True
    )

    SENTINEL1 = DataCollectionDefinition(
        api_id='sentinel-1-grd',
        catalog_id='sentinel-1-grd',
        wfs_id='DSS3',
        collection_type=_CollectionType.SENTINEL1,
        sensor_type=_SensorType.C_SAR,
        processing_level=_ProcessingLevel.GRD,
        orbit_direction=OrbitDirection.BOTH
    )
    SENTINEL1_IW = SENTINEL1.derive(
        swath_mode=_SwathMode.IW,
        polarization=_Polarization.DV,
        resolution=_Resolution.HIGH,
        bands=_Bands.SENTINEL1_IW
    )
    SENTINEL1_IW_ASC = SENTINEL1_IW.derive(
        orbit_direction=OrbitDirection.ASCENDING
    )
    SENTINEL1_IW_DES = SENTINEL1_IW.derive(
        orbit_direction=OrbitDirection.DESCENDING
    )
    SENTINEL1_EW = SENTINEL1.derive(
        swath_mode=_SwathMode.EW,
        polarization=_Polarization.DH,
        resolution=_Resolution.MEDIUM,
        bands=_Bands.SENTINEL1_EW
    )
    SENTINEL1_EW_ASC = SENTINEL1_EW.derive(
        orbit_direction=OrbitDirection.ASCENDING
    )
    SENTINEL1_EW_DES = SENTINEL1_EW.derive(
        orbit_direction=OrbitDirection.DESCENDING
    )
    SENTINEL1_EW_SH = SENTINEL1_EW.derive(
        polarization=_Polarization.SH,
        bands=_Bands.SENTINEL1_EW_SH
    )
    SENTINEL1_EW_SH_ASC = SENTINEL1_EW_SH.derive(
        orbit_direction=OrbitDirection.ASCENDING
    )
    SENTINEL1_EW_SH_DES = SENTINEL1_EW_SH.derive(
        orbit_direction=OrbitDirection.DESCENDING
    )

    DEM = DataCollectionDefinition(
        api_id='dem',
        collection_type=_CollectionType.DEM,
        bands=_Bands.DEM,
        is_timeless=True
    )
    DEM_MAPZEN = DEM.derive(
        dem_instance='MAPZEN'
    )
    DEM_COPERNICUS_30 = DEM.derive(
        dem_instance='COPERNICUS_30'
    )
    DEM_COPERNICUS_90 = DEM.derive(
        dem_instance='COPERNICUS_90'
    )

    MODIS = DataCollectionDefinition(
        api_id='modis',
        catalog_id='modis',
        wfs_id='DSS5',
        service_url=ServiceUrl.USWEST,
        collection_type=_CollectionType.MODIS,
        bands=_Bands.MODIS
    )

    LANDSAT15_L1 = DataCollectionDefinition(
        api_id='landsat-mss-l1',
        catalog_id='landsat-mss-l1',
        wfs_id='DSS14',
        service_url=ServiceUrl.USWEST,
        collection_type=_CollectionType.LANDSAT15,
        sensor_type=_SensorType.MSS,
        processing_level=_ProcessingLevel.L1,
        bands=_Bands.LANDSAT15,
        has_cloud_coverage=True
    )

    LANDSAT45_L1 = DataCollectionDefinition(
        api_id='landsat-tm-l1',
        catalog_id='landsat-tm-l1',
        wfs_id='DSS15',
        service_url=ServiceUrl.USWEST,
        collection_type=_CollectionType.LANDSAT45,
        sensor_type=_SensorType.TM,
        processing_level=_ProcessingLevel.L1,
        bands=_Bands.LANDSAT45,
        has_cloud_coverage=True
    )
    LANDSAT45_L2 = LANDSAT45_L1.derive(
        api_id='landsat-tm-l2',
        catalog_id='landsat-tm-l2',
        wfs_id='DSS16',
        processing_level=_ProcessingLevel.L2
    )

    LANDSAT7_L1 = DataCollectionDefinition(
        api_id='landsat-etm-l1',
        catalog_id='landsat-etm-l1',
        wfs_id='DSS17',
        service_url=ServiceUrl.USWEST,
        collection_type=_CollectionType.LANDSAT7,
        sensor_type=_SensorType.ETM,
        processing_level=_ProcessingLevel.L1,
        bands=_Bands.LANDSAT7_L1,
        has_cloud_coverage=True
    )
    LANDSAT7_L2 = LANDSAT7_L1.derive(
        api_id='landsat-etm-l2',
        catalog_id='landsat-etm-l2',
        wfs_id='DSS18',
        processing_level=_ProcessingLevel.L2,
        bands=_Bands.LANDSAT7_L2
    )

    LANDSAT8 = DataCollectionDefinition(
        api_id='landsat-8-l1c',
        catalog_id='landsat-8-l1c',
        wfs_id='DSS6',
        service_url=ServiceUrl.USWEST,
        collection_type=_CollectionType.LANDSAT8,
        sensor_type=_SensorType.OLI_TIRS,
        processing_level=_ProcessingLevel.L1,
        bands=_Bands.LANDSAT8,
        has_cloud_coverage=True
    )
    LANDSAT8_L1 = LANDSAT8.derive(
        api_id='landsat-ot-l1',
        catalog_id='landsat-ot-l1',
        wfs_id='DSS12'
    )
    LANDSAT8_L2 = LANDSAT8.derive(
        api_id='landsat-ot-l2',
        catalog_id='landsat-ot-l2',
        wfs_id='DSS13',
        processing_level=_ProcessingLevel.L2,
        bands=_Bands.LANDSAT8_L2
    )

    SENTINEL5P = DataCollectionDefinition(
        api_id='sentinel-5p-l2',
        catalog_id='sentinel-5p-l2',
        wfs_id='DSS7',
        service_url=ServiceUrl.CREODIAS,
        collection_type=_CollectionType.SENTINEL5P,
        sensor_type=_SensorType.TROPOMI,
        processing_level=_ProcessingLevel.L2,
        bands=_Bands.SENTINEL5P
    )
    SENTINEL3_OLCI = DataCollectionDefinition(
        api_id='sentinel-3-olci',
        catalog_id='sentinel-3-olci',
        wfs_id='DSS8',
        service_url=ServiceUrl.CREODIAS,
        collection_type=_CollectionType.SENTINEL3,
        sensor_type=_SensorType.OLCI,
        processing_level=_ProcessingLevel.L1B,
        bands=_Bands.SENTINEL3_OLCI
    )
    SENTINEL3_SLSTR = DataCollectionDefinition(
        api_id='sentinel-3-slstr',
        catalog_id='sentinel-3-slstr',
        wfs_id='DSS9',
        service_url=ServiceUrl.CREODIAS,
        collection_type=_CollectionType.SENTINEL3,
        sensor_type=_SensorType.SLSTR,
        processing_level=_ProcessingLevel.L1B,
        bands=_Bands.SENTINEL3_SLSTR,
        has_cloud_coverage=True
    )

    # EOCloud collections (which are only available on a development eocloud service):
    LANDSAT5 = DataCollectionDefinition(
        wfs_id='L5.TILE',
        service_url=ServiceUrl.EOCLOUD,
        processing_level=_ProcessingLevel.GRD
    )
    LANDSAT7 = DataCollectionDefinition(
        wfs_id='L7.TILE',
        service_url=ServiceUrl.EOCLOUD,
        processing_level=_ProcessingLevel.GRD
    )
    ENVISAT_MERIS = DataCollectionDefinition(
        wfs_id='ENV.TILE',
        service_url=ServiceUrl.EOCLOUD,
        collection_type=_CollectionType.ENVISAT_MERIS
    )

    @classmethod
    def define(cls, name, *, api_id=None, catalog_id=None, wfs_id=None, service_url=None, collection_type=None,
               sensor_type=None, processing_level=None, swath_mode=None, polarization=None, resolution=None,
               orbit_direction=None, timeliness=None, bands=None, collection_id=None, is_timeless=False,
               has_cloud_coverage=False, dem_instance=None):
        """ Define a new data collection

        Note that all parameters, except `name` are optional. If a data collection definition won't be used for a
        certain use case (e.g. Process API, WFS, etc.), parameters for that use case don't have to be defined

        :param name: A name of a new data collection
        :type name: str
        :param api_id: An ID to be used for Sentinel Hub Process API
        :type api_id: str or None
        :param catalog_id: An ID to be used for Sentinel Hub Catalog API
        :type catalog_id: str or None
        :param wfs_id: An ID to be used for Sentinel Hub WFS service
        :type wfs_id: str or None
        :param service_url: A base URL of Sentinel Hub service deployment from where to download data. If it is not
            specified, a `sh_base_url` from a config will be used by default
        :type service_url: str or None
        :param collection_type: A collection type
        :type collection_type: str or None
        :param sensor_type: A type of a satellite's sensor
        :type sensor_type: str or None
        :param processing_level: A level of processing applied on satellite data
        :type processing_level: str or None
        :param swath_mode: A swath mode of SAR sensors
        :type swath_mode: str or None
        :param polarization: A type of polarization
        :type polarization: str or None
        :param resolution: A type of (Sentinel-1) resolution
        :type resolution: str or None
        :param orbit_direction: A direction of satellite's orbit by which to filter satellite's data
        :type orbit_direction: str or None
        :param timeliness: A timeliness of data
        :type timeliness: str or None
        :param bands: Names of data collection bands
        :type bands: tuple(str) or None
        :param collection_id: An ID of a BYOC or BATCH collection
        :type collection_id: str or None
        :param is_timeless: `True` if a data collection can be filtered by time dimension and `False` otherwise
        :type is_timeless: bool
        :param has_cloud_coverage: `True` if data collection can be filtered by cloud coverage percentage and `False`
            otherwise
        :type has_cloud_coverage: bool
        :param dem_instance: one of the options listed in
            `DEM documentation <https://docs.sentinel-hub.com/api/latest/data/dem/#deminstance>`__
        :type dem_instance: str or None
        :return: A new data collection
        :rtype: DataCollection
        """
        definition = DataCollectionDefinition(
            api_id=api_id,
            catalog_id=catalog_id,
            wfs_id=wfs_id,
            service_url=service_url,
            collection_type=collection_type,
            sensor_type=sensor_type,
            processing_level=processing_level,
            swath_mode=swath_mode,
            polarization=polarization,
            resolution=resolution,
            orbit_direction=orbit_direction,
            timeliness=timeliness,
            bands=bands,
            collection_id=collection_id,
            is_timeless=is_timeless,
            has_cloud_coverage=has_cloud_coverage,
            dem_instance=dem_instance,
            _name=name
        )
        cls._try_add_data_collection(name, definition)
        return cls(definition)

    def define_from(self, name, **params):
        """ Define a new data collection from an existing one

        :param name: A name of a new data collection
        :type name: str
        :param params: Any parameter to override current data collection parameters
        :return: A new data collection
        :rtype: DataCollection
        """
        definition = self.value
        new_definition = definition.derive(**params, _name=name)

        self._try_add_data_collection(name, new_definition)
        return DataCollection(new_definition)

    @classmethod
    def _try_add_data_collection(cls, name, definition):
        """ Tries adding a new data collection definition. If the exact enum has already been defined then it won't do
        anything. However, if either a name or a definition has already been matched with another name or definition
        then it will raise an error.
        """
        is_name_defined = name in cls.__members__
        is_enum_defined = is_name_defined and cls.__members__[name].value == definition
        is_definition_defined = definition in cls._value2member_map_

        if is_enum_defined:
            return

        if not is_name_defined and not is_definition_defined:
            extend_enum(cls, name, definition)
            return

        if is_name_defined:
            raise ValueError(f"Data collection name '{name}' is already taken by another data collection")

        existing_collection = cls._value2member_map_[definition]
        raise ValueError(f'Data collection definition is already taken by {existing_collection}. Two different '
                         f'DataCollection enums cannot have the same definition.')

    @classmethod
    def define_byoc(cls, collection_id, **params):
        """ Defines a BYOC data collection

        :param collection_id: An ID of a data collection
        :type collection_id: str
        :param params: Any parameter to override default BYOC data collection parameters
        :return: A new data collection
        :rtype: DataCollection
        """
        params['name'] = params.get('name', f'BYOC_{collection_id}')
        params['api_id'] = params.get('api_id', f'byoc-{collection_id}')
        params['catalog_id'] = params.get('catalog_id', f'byoc-{collection_id}')
        params['wfs_id'] = params.get('wfs_id', f'byoc-{collection_id}')
        params['collection_type'] = params.get('collection_type', _CollectionType.BYOC)
        params['collection_id'] = collection_id
        return cls.define(**params)

    @classmethod
    def define_batch(cls, collection_id, **params):
        """ Defines a BATCH data collection

        :param collection_id: An ID of a data collection
        :type collection_id: str
        :param params: Any parameter to override default BATCH data collection parameters
        :return: A new data collection
        :rtype: DataCollection
        """
        params['name'] = params.get('name', f'BATCH_{collection_id}')
        params['api_id'] = params.get('api_id', f'batch-{collection_id}')
        params['catalog_id'] = params.get('catalog_id', f'batch-{collection_id}')
        params['wfs_id'] = params.get('wfs_id', f'batch-{collection_id}')
        params['collection_type'] = params.get('collection_type', _CollectionType.BATCH)
        params['collection_id'] = collection_id
        return cls.define(**params)

    @property
    def api_id(self):
        """ Provides a Sentinel Hub Process API identifier or raises an error if it is not defined

        :return: An identifier
        :rtype: str
        :raises: ValueError
        """
        if self.value.api_id is None:
            raise ValueError(f'Data collection {self.name} is missing a Sentinel Hub Process API identifier')
        return self.value.api_id

    @property
    def catalog_id(self):
        """ Provides a Sentinel Hub Catalog API identifier or raises an error if it is not defined

        :return: An identifier
        :rtype: str
        :raises: ValueError
        """
        if self.value.catalog_id is None:
            raise ValueError(f'Data collection {self.name} is missing a Sentinel Hub Catalog API identifier')
        return self.value.catalog_id

    @property
    def wfs_id(self):
        """ Provides a Sentinel Hub WFS identifier or raises an error if it is not defined

        :return: An identifier
        :rtype: str
        :raises: ValueError
        """
        if self.value.wfs_id is None:
            raise ValueError(f'Data collection {self.name} is missing a Sentinel Hub WFS identifier')
        return self.value.wfs_id

    @property
    def bands(self):
        """ Provides band names available for the data collection

        :return: A tuple of band names
        :rtype: tuple(str)
        :raises: ValueError
        """
        if self.value.bands is None:
            raise ValueError(f'Data collection {self.name} does not define bands')
        return self.value.bands

    def __getattr__(self, item, *args, **kwargs):
        """ The following insures that any attribute from DataCollectionDefinition, which is already not a
        property or an attribute of DataCollection, becomes an attribute of DataCollection
        """
        if not item.startswith('_') and hasattr(self, 'value') and isinstance(self.value, DataCollectionDefinition):
            definition_dict = asdict(self.value)
            if item in definition_dict:
                return definition_dict[item]

        return super().__getattribute__(item, *args, **kwargs)

    @property
    def is_sentinel1(self):
        """ Checks if data collection is a Sentinel-1 collection type

        Example: ``DataCollection.SENTINEL1_IW.is_sentinel1``

        :return: `True` if collection is Sentinel-1 collection type and `False` otherwise
        :rtype: bool
        """
        return self.collection_type == _CollectionType.SENTINEL1

    def contains_orbit_direction(self, orbit_direction):
        """ Checks if a data collection contains given orbit direction

        :param orbit_direction: An orbit direction
        :type orbit_direction: string
        :return: `True` if data collection contains the orbit direction
        :return: bool
        """
        defined_direction = self.orbit_direction
        if defined_direction is None or defined_direction.upper() == OrbitDirection.BOTH:
            return True
        return orbit_direction.upper() == defined_direction.upper()

    @classmethod
    def get_available_collections(cls, config=None):
        """ Returns which data collections are available for configured Sentinel Hub OGC URL

        :param config: A custom instance of config class to override parameters from the saved configuration.
        :type config: SHConfig or None
        :return: List of available data collections
        :rtype: list(DataCollection)
        """
        config = config or SHConfig()
        is_eocloud = config.has_eocloud_url()

        return [data_collection for data_collection in cls
                if (data_collection.service_url == ServiceUrl.EOCLOUD) == is_eocloud]


def _raise_invalid_id(collection_id):
    """ Checks if a given collection ID conforms to an expected pattern and raises an error if it doesn't
    """
    collection_id_pattern = '.{8}-.{4}-.{4}-.{4}-.{12}'

    if not re.compile(collection_id_pattern).match(collection_id):
        raise ValueError("Given collection id does not match the expected format {}".format(collection_id_pattern))


DataSource = DataCollection


def handle_deprecated_data_source(data_collection, data_source, default=None):
    """ Joins parameters used to specify a data collection. In case data_source is given it raises a warning. In case
    both are given it raises an error. In case neither are given but there is a default collection it raises another
    warning.

    Note that this function is only temporary and will be removed in future package versions
    """
    if data_source is not None:
        warnings.warn('Parameter data_source is deprecated, use data_collection instead',
                      category=SHDeprecationWarning)

    if data_collection is not None and data_source is not None:
        raise ValueError('Only one of the parameters data_collection and data_source should be given')

    if data_collection is None and data_source is None and default is not None:
        warnings.warn('In the future please specify data_collection parameter, for now taking '
                      'DataCollection.SENTINEL2_L1C', category=SHDeprecationWarning)
        return default

    return data_collection or data_source



"""
Module implementing DownloadRequest class
"""
import hashlib
import json
import os
import warnings
import datetime as dt

from ..constants import MimeType, RequestType
from ..exceptions import SHRuntimeWarning
from ..os_utils import sys_is_windows


class DownloadRequest:
    """ A class defining a single download request.

    The class is a container with all parameters needed to execute a single download request and save or return
    downloaded data.
    """
    def __init__(self, *, url=None, headers=None, request_type=RequestType.GET, post_values=None, use_session=False,
                 data_type=MimeType.RAW, save_response=False, data_folder=None, filename=None, return_data=True,
                 **properties):
        """
        :param url: An URL from where to download
        :type url: str or None
        :param headers: Headers of a HTTP request
        :type headers: dict or None
        :param request_type: Type of request, either GET or POST. Default is `RequestType.GET`
        :type request_type: str or RequestType
        :param post_values: A dictionary of values that will be sent with a POST request. Default is `None`
        :type post_values: dict or None
        :param use_session: A flag that specifies if the download request will require a session header
        :type use_session: bool
        :param data_type: An expected file format of downloaded data. Default is `MimeType.RAW`
        :type data_type: MimeType or str
        :param save_response: A flag defining if the downloaded data will be saved to disk. Default is `False`.
        :type save_response: bool
        :param data_folder: A folder path where the fetched data will be (or already is) saved. Default is `None`
        :type data_folder: str or None
        :param filename: A custom filename where the data will be saved. By default data will be saved in a folder
            which name are hashed request parameters.
        :type filename: str or None
        :param return_data: A flag defining if the downloaded data will be returned as an output of download procedure.
            Default is `True`.
        :type return_data: bool
        :param properties: Any additional parameters.
        """
        self.url = url
        self.headers = headers or {}
        self.request_type = RequestType(request_type)
        self.post_values = post_values
        self.use_session = use_session

        self.data_type = MimeType(data_type)

        self.save_response = save_response
        self.data_folder = data_folder
        self.filename = filename
        self.return_data = return_data

        self.properties = properties

    def raise_if_invalid(self):
        """ Method that raises an error if something is wrong with request parameters

        :raises: ValueError
        """
        if self.save_response and self.data_folder is None:
            raise ValueError('Data folder is not specified. '
                             'Please give a data folder name in the initialization of your request.')

    def get_request_params(self, include_metadata=False):
        """ Provides parameters that define the request in form of a dictionary

        :param include_metadata: A flag defining if also metadata parameters should be included, such as headers and
            current time
        :type include_metadata: bool
        :return: A dictionary of parameters
        :rtype: dict
        """
        params = {
            'url': self.url,
            'payload': self.post_values
        }
        if include_metadata:
            params = {
                **params,
                'headers': self.headers,
                'timestamp': dt.datetime.now().isoformat()
            }
        return params

    def get_hashed_name(self):
        """ It takes request url and payload and calculates a unique hashed string from them.

        :return: A hashed string
        :rtype: str
        """
        params = self.get_request_params(include_metadata=False)
        hashable = json.dumps(params)

        return hashlib.md5(hashable.encode('utf-8')).hexdigest()

    def get_relative_paths(self):
        """ A method that calculates file paths relative to `data_folder`

        :return: Returns a pair of file paths, a request payload path and a response path. If request path is not
            defined it returns `None`.
        :rtype: (str or None, str)
        """

        if self.filename is not None:
            return None, self.filename

        hashed_name = self.get_hashed_name()

        request_path = os.path.join(hashed_name, 'request.json')
        response_path = os.path.join(hashed_name, 'response.{}'.format(self.data_type.extension))

        return request_path, response_path

    def get_storage_paths(self):
        """ A method that calculates file paths where request payload and response will be saved.

        :return: Returns a pair of file paths, a request payload path and a response path. Each of them can also be
            `None` if it is not defined.
        :rtype: (str or None, str or None)
        """
        if self.data_folder is None:
            return None, None

        request_path, response_path = self.get_relative_paths()

        if request_path is not None:
            request_path = os.path.join(self.data_folder, request_path)
        response_path = os.path.join(self.data_folder, response_path)

        self._check_path(response_path)
        return request_path, response_path

    @staticmethod
    def _check_path(file_path):
        """ Checks file path and warns about potential problems during saving
        """
        message_problem = None
        if len(file_path) > 255 and sys_is_windows():
            message_problem = 'File path'
        elif len(os.path.basename(file_path)) > 255:
            message_problem = 'Filename of'

        if message_problem:
            message = '{} {} is longer than 255 character which might cause an error while saving on ' \
                      'disk'.format(message_problem, file_path)
            warnings.warn(message, category=SHRuntimeWarning)
